from protowhat.Feedback import FeedbackComponent
from protowhat.utils_messaging import get_times
from tcs_pythonwhat.Test import BiggerTest


def test_object_accessed(state, name, times=1, not_accessed_msg=None):
    """Test if object accessed

    Checks whether an object, or the attribute of an object, are accessed

    Args:
        name (str): the name of the object that should be accessed; can contain dots (for attributes)
        times (int): how often the object specified in name should be accessed.
        not_accessed_msg (str): custom feedback message when the object was not accessed.

    Examples:


        Student code

        | ``import numpy as np``
        | ``arr = np.array([1, 2, 3])``
        | ``x = arr.shape``

        Solution code

        | ``import numpy as np``
        | ``arr = np.array([1, 2, 3])``
        | ``x = arr.shape``
        | ``t = arr.dtype``

        SCT

        | ``test_object_accessed("arr")``: pass.
        | ``test_object_accessed("arr.shape")``: pass.
        | ``test_object_accessed("arr.dtype")``: fail.
    """
    student_object_accesses = state.ast_dispatcher.find(
        "object_accesses", state.student_ast
    )
    student_mappings = state.ast_dispatcher.find("oa_mappings", state.student_ast)

    if not not_accessed_msg:
        stud_name = name
        if "." in stud_name:
            for orig, full_name in student_mappings.items():
                if name.startswith(full_name):
                    stud_name = name.replace(full_name, orig)

        add = " at least %s" % get_times(times) if times > 1 else ""
        not_accessed_msg = "Have you accessed `%s`%s?" % (stud_name, add)

    # name should be contained inside the student_object_accesses.
    # hack: add a dot and do a match on the name with the dot,
    # to make sure you're not matching substrings
    student_hits = [c for c in student_object_accesses if name + "." in c + "."]
    state.do_test(
        BiggerTest(len(student_hits) + 1, times, FeedbackComponent(not_accessed_msg))
    )



from protowhat.Feedback import FeedbackComponent
from protowhat.checks.check_logic import (
    multi,
    check_not,
    check_or,
    check_correct,
    disable_highlighting,
    fail,
)
from protowhat.failure import InstructorError
import ast


multi.__doc__ = (
    str(multi.__doc__)
    + """
    :Example:

        Suppose we want to verify the following function call: ::

            round(1.2345, ndigits=2)

        The following SCT would verify this, using ``multi`` to
        'branch out' the state to two sub-SCTs: ::

            Ex().check_function('round').multi(
                check_args(0).has_equal_value(),
                check_args('ndigits').has_equal_value()
            )
    """
)


check_not.__doc__ = (
    str(check_not.__doc__)
    + """
    :Example:
        The SCT fails with feedback for a specific incorrect value, defined using an override: ::

            Ex().check_object('result').multi(
                check_not(
                    has_equal_value(override=100),
                    msg='100 is incorrect for reason xyz.'
                ),
                has_equal_value()
            )

        Notice that ``check_not`` comes before the ``has_equal_value`` test
        that checks if the student value is equal to the solution value.

    :Example:
        The SCT below runs two ``has_code`` cases: ::

            Ex().check_not(
                has_code('mean'),
                has_code('median'),
                msg='Check your code'
            )

        If students use ``mean`` or ``median`` anywhere in their code, this SCT will fail.

    Note:
        - This function is not yet tested with all checks, please report unexpected behaviour.
        - This function can be thought as a NOT(x OR y OR ...) statement, since all tests it runs must fail
        - This function can be considered a direct counterpart of multi.

    """
)


check_or.__doc__ = (
    str(check_or.__doc__)
    + """
    :Example:

        The SCT below tests that the student typed either 'mean' or 'median': ::

            Ex().check_or(
                has_code('mean'),
                has_code('median')
            )

        If the student didn't type either, the feedback message generated by ``has_code(mean)``,
        the first SCT, will be presented to the student.

    """
)


check_correct.__doc__ = (
    str(check_correct.__doc__)
    + """
    :Example:

        The SCT below tests whether an object is correct. Only if the object is not correct, will
        the function calling checks be executed ::

            Ex().check_correct(
                check_object('x').has_equal_value(),
                check_function('round').check_args(0).has_equal_value()
            )

    """
)


# utility functions -----------------------------------------------------------


fail.__doc__ = (
    str(fail.__doc__)
    + """
    :Example:

        As a trivial SCT example, ::

            Ex().check_for_loop().check_body().fail()

        This can also be helpful for debugging SCTs, as it can be used to stop testing as a given point.
    """
)


def override(state, solution):
    """Override the solution code with something arbitrary.

    There might be cases in which you want to temporarily override the solution code
    so you can allow for alternative ways of solving an exercise.
    When you use ``override()`` in an SCT chain, the remainder of that SCT chain will
    run as if the solution code you specified is the only code that was in the solution.

    Check the glossary for an example (pandas plotting)

    Args:
        solution: solution code as a string that overrides the original solution code.
        state: State instance describing student and solution code. Can be omitted if used with Ex().
    """

    # the old ast may be a number of node types, but generally either a
    # (1) ast.Module, or for single expressions...
    # (2) whatever was grabbed using module.body[0]
    # (3) module.body[0].value, when module.body[0] is an Expr node
    old_ast = state.solution_ast
    new_ast = ast.parse(solution)
    if not isinstance(old_ast, ast.Module) and len(new_ast.body) == 1:
        expr = new_ast.body[0]
        candidates = [expr, expr.value] if isinstance(expr, ast.Expr) else [expr]
        for node in candidates:
            if isinstance(node, old_ast.__class__):
                new_ast = node
                break

    kwargs = state.feedback_context.kwargs if state.feedback_context else {}
    child = state.to_child(
        solution_ast=new_ast,
        student_ast=state.student_ast,
        highlight=state.highlight,
        append_message=FeedbackComponent("", kwargs),
    )

    return child


def set_context(state, *args, **kwargs):
    """Update context values for student and solution environments.

    When ``has_equal_x()`` is used after this, the context values (in ``for`` loops and function definitions, for example)
    will have the values specified through his function. It is the function equivalent of the ``context_vals`` argument of
    the ``has_equal_x()`` functions.

    - Note 1: excess args and unmatched kwargs will be unused in the student environment.
    - Note 2: When you try to set context values that don't match any target variables in the solution code,
      ``set_context()`` raises an exception that lists the ones available.
    - Note 3: positional arguments are more robust to the student using different names for context values.
    - Note 4: You have to specify arguments either by position, either by name. A combination is not possible.

    :Example:

        Solution code::

            total = 0
            for i in range(10):
                print(i ** 2)

        Student submission that will pass (different iterator, different calculation)::

            total = 0
            for j in range(10):
                print(j * j)

        SCT::

            # set_context is robust against different names of context values.
            Ex().check_for_loop().check_body().multi(
                set_context(1).has_equal_output(),
                set_context(2).has_equal_output(),
                set_context(3).has_equal_output()
            )

            # equivalent SCT, by setting context_vals in has_equal_output()
            Ex().check_for_loop().check_body().\\
                multi([s.has_equal_output(context_vals=[i]) for i in range(1, 4)])

    """

    stu_crnt = state.student_context.context
    sol_crnt = state.solution_context.context

    # for now, you can't specify both
    if len(args) > 0 and len(kwargs) > 0:
        raise InstructorError.from_message(
            "В `set_context()`, укажите аргументы либо позиционно, либо по ключу."
        )

    # set args specified by pos -----------------------------------------------
    if args:
        # stop if too many pos args for solution
        if len(args) > len(sol_crnt):
            raise InstructorError.from_message(
                "Слишком много позиционных аргументов. Существуют контекстные переменные {}, но вы попытались передать {}".format(
                    len(sol_crnt), len(args)
                )
            )
        # set pos args
        upd_sol = sol_crnt.update(dict(zip(sol_crnt.keys(), args)))
        upd_stu = stu_crnt.update(dict(zip(stu_crnt.keys(), args)))
    else:
        upd_sol = sol_crnt
        upd_stu = stu_crnt

    # set args specified by keyword -------------------------------------------
    if kwargs:
        # stop if keywords don't match with solution
        if set(kwargs) - set(upd_sol):
            raise InstructorError.from_message(
                "`set_context()`: контекстные переменные называются {}, а вы попытались передать {}.".format(
                    upd_sol or "missing", sorted(list(kwargs.keys()))
                )
            )
        out_sol = upd_sol.update(kwargs)
        # need to match keys in kwargs with corresponding keys in stu context
        # in case they used, e.g., different loop variable names
        match_keys = dict(zip(sol_crnt.keys(), stu_crnt.keys()))
        out_stu = upd_stu.update(
            {match_keys[k]: v for k, v in kwargs.items() if k in match_keys}
        )
    else:
        out_sol = upd_sol
        out_stu = upd_stu

    return state.to_child(
        student_context=out_stu, solution_context=out_sol, highlight=state.highlight
    )


def set_env(state, **kwargs):
    """Update/set environemnt variables for student and solution environments.

    When ``has_equal_x()`` is used after this, the variables specified through this function will
    be available in the student and solution process. Note that you will not see these variables
    in the student process of the state produced by this function: the values are saved on the state
    and are only added to the student and solution processes when ``has_equal_ast()`` is called.

    :Example:

        Student and Solution Code::

            a = 1
            if a > 4:
                print('pretty large')

        SCT::

            # check if condition works with different values of a
            Ex().check_if_else().check_test().multi(
                set_env(a = 3).has_equal_value(),
                set_env(a = 4).has_equal_value(),
                set_env(a = 5).has_equal_value()
            )

            # equivalent SCT, by setting extra_env in has_equal_value()
            Ex().check_if_else().check_test().\\
                multi([has_equal_value(extra_env={'a': i}) for i in range(3, 6)])
    """

    stu_crnt = state.student_env.context
    sol_crnt = state.solution_env.context

    stu_new = stu_crnt.update(kwargs)
    sol_new = sol_crnt.update(kwargs)

    return state.to_child(
        student_env=stu_new, solution_env=sol_new, highlight=state.highlight
    )


disable_highlighting.__doc__ = (
    str(disable_highlighting.__doc__)
    + """
    :Examples:

        SCT that will mark the 'number' portion if it is incorrect::

            Ex().check_function('round').check_args(0).has_equal_ast()

        SCT chains that will not mark certain mistakes. The earlier you put the function, the more types of mistakes will no longer be highlighted::

            Ex().disable_highlighting().check_function('round').check_args(0).has_equal_ast()
            Ex().check_function('round').disable_highlighting().check_args(0).has_equal_ast()
            Ex().check_function('round').check_args(0).disable_highlighting().has_equal_ast()
    """
)



from tcs_pythonwhat.Test import EqualTest
from protowhat.Feedback import FeedbackComponent
from protowhat.failure import debugger
from tcs_pythonwhat.State import State
from functools import singledispatch
from tcs_pythonwhat.checks.check_funcs import check_part_index

MSG_INCORRECT_LOOP = "Вы использовали корректные имена переменных итератора? Ожидалось `{{sol_vars}}`, но было получено `{{stu_vars}}`."
MSG_INCORRECT_WITH = "Убедитесь, что вы использовали корректные контекстные переменные. Ожидалось `{{sol_vars}}`, но было получено `{{stu_vars}}`."


def has_context(state, incorrect_msg=None, exact_names=False):
    # call _has_context, since the built-in singledispatch can only use 1st pos arg
    return _has_context(state, incorrect_msg, exact_names)


def _test(state, incorrect_msg, exact_names, tv_name, highlight_name):
    # get parts for testing from state
    # TODO: this could be rewritten to use check_part_index -> has_equal_part, etc..
    stu_vars = state.student_parts[tv_name]
    sol_vars = state.solution_parts[tv_name]

    child_state = state.to_child(
        student_ast=state.student_parts.get(highlight_name),
        solution_ast=state.solution_parts.get(highlight_name),
    )

    # variables exposed to messages
    d = {"stu_vars": stu_vars, "sol_vars": sol_vars, "num_vars": len(sol_vars)}

    if exact_names:
        # feedback for wrong iter var names
        child_state.do_test(
            EqualTest(stu_vars, sol_vars, FeedbackComponent(incorrect_msg, d))
        )
    else:
        # feedback for wrong number of iter vars
        child_state.do_test(
            EqualTest(len(stu_vars), len(sol_vars), FeedbackComponent(incorrect_msg, d))
        )

    return state


@singledispatch
def _has_context(state, incorrect_msg, exact_names):
    with debugger(state):
        state.report(
            "первым аргументом _has_context должен быть экземпляр State или его подкласс"
        )


@_has_context.register(State)
def has_context_state(*args, **kwargs):
    return _test(*args, tv_name="target_vars", highlight_name="highlight", **kwargs)


@_has_context.register(State.SUBCLASSES["for_loops"])
@_has_context.register(State.SUBCLASSES["whiles"])
@_has_context.register(State.SUBCLASSES["dict_comps"])
@_has_context.register(State.SUBCLASSES["generator_exps"])
@_has_context.register(State.SUBCLASSES["list_comps"])
def has_context_loop(state, incorrect_msg, exact_names):
    """When dispatched on loops, has_context the target vars are the attribute _target_vars.

    Note: This is to allow people to call has_context on a node (e.g. for_loop) rather than
          one of its attributes (e.g. body). Purely for convenience.
    """
    return _test(
        state,
        incorrect_msg or MSG_INCORRECT_LOOP,
        exact_names,
        tv_name="_target_vars",
        highlight_name="target",
    )


@_has_context.register(State.SUBCLASSES["withs"])
def has_context_with(state, incorrect_msg, exact_names):
    """When dispatched on with statements, has_context loops over each context manager.

    Note: This is to allow people to call has_context on the with statement, rather than
          having to manually loop over each context manager.

          e.g. Ex().check_with(0).has_context() vs Ex().check_with(0).check_context(0).has_context()
    """

    for i in range(len(state.solution_parts["context"])):
        ctxt_state = check_part_index(state, "context", i, "{{ordinal}} context")
        _has_context(ctxt_state, incorrect_msg or MSG_INCORRECT_WITH, exact_names)

    return state



import ast

from protowhat.Feedback import FeedbackComponent
from protowhat.failure import InstructorError, debugger
from tcs_pythonwhat.Test import EqualTest
from tcs_pythonwhat.checks.has_funcs import evalCalls
from tcs_pythonwhat.tasks import ReprFail


def fix_format(arguments):
    if isinstance(arguments, str):
        arguments = (arguments,)
    if isinstance(arguments, tuple):
        arguments = list(arguments)

    if isinstance(arguments, list):
        arguments = {"args": arguments, "kwargs": {}}

    if (
        not isinstance(arguments, dict)
        or "args" not in arguments
        or "kwargs" not in arguments
    ):
        raise ValueError(
            "Wrong format of arguments in 'results', 'outputs' or 'errors'; either a list, or a dictionary with names args (a list) and kwargs (a dict)"
        )

    return arguments


def stringify(arguments):
    vararg = str(arguments["args"])[1:-1]
    kwarg = ", ".join(
        ["%s = %s" % (key, value) for key, value in arguments["kwargs"].items()]
    )
    if len(vararg) == 0:
        if len(kwarg) == 0:
            return "()"
        else:
            return "(" + kwarg + ")"
    else:
        if len(kwarg) == 0:
            return "(" + vararg + ")"
        else:
            return "(" + ", ".join([vararg, kwarg]) + ")"


# TODO: test string syntax with check_function_def
#       test argument syntax with check_lambda_function
def run_call(args, node, process, get_func, **kwargs):
    # Get function expression
    if isinstance(node, ast.FunctionDef):  # function name
        func_expr = ast.Name(id=node.name, ctx=ast.Load())
    elif isinstance(node, ast.Lambda):  # lambda body expr
        func_expr = node
    else:
        raise InstructorError.from_message(
            "Only function definition or lambda may be called"
        )

    ast.fix_missing_locations(func_expr)
    return get_func(process=process, tree=func_expr, call=args, **kwargs)


MSG_CALL_INCORRECT = "Calling {{argstr}} should {{action}} `{{str_sol}}`, instead got {{str_stu if str_stu == 'no printouts' else '`' + str_stu + '`'}}."
MSG_CALL_ERROR = "Calling {{argstr}} should {{action}} `{{str_sol}}`, instead it errored out: `{{str_stu}}`."
MSG_CALL_ERROR_INV = (
    "Calling {{argstr}} should {{action}} `{{str_sol}}`, instead got `{{str_stu}}`."
)


def call(
    state,
    args,
    test="value",
    incorrect_msg=None,
    error_msg=None,
    argstr=None,
    func=None,
    **kwargs
):
    """Use ``check_call()`` in combination with ``has_equal_x()`` instead.
    """

    if incorrect_msg is None:
        incorrect_msg = MSG_CALL_INCORRECT
    if error_msg is None:
        error_msg = MSG_CALL_ERROR_INV if test == "error" else MSG_CALL_ERROR

    assert test in ("value", "output", "error")

    get_func = evalCalls[test]

    # Run for Solution --------------------------------------------------------
    eval_sol, str_sol = run_call(
        args, state.solution_parts["node"], state.solution_process, get_func, **kwargs
    )

    if (test == "error") ^ isinstance(eval_sol, Exception):
        with debugger(state):
            state.report(
                "Calling {{argstr}} resulted in an error (or not an error if testing for one). Error message: {{type_err}} {{str_sol}}",
                dict(type_err=type(eval_sol), str_sol=str_sol, argstr=argstr),
            )

    if isinstance(eval_sol, ReprFail):
        with debugger(state):
            state.report(
                "Can't get the result of calling {{argstr}}: {{eval_sol.info}}",
                dict(argstr=argstr, eval_sol=eval_sol),
            )

    # Run for Submission ------------------------------------------------------
    eval_stu, str_stu = run_call(
        args, state.student_parts["node"], state.student_process, get_func, **kwargs
    )
    action_strs = {
        "value": "return",
        "output": "print out",
        "error": "error out with the message",
    }
    fmt_kwargs = {
        "part": argstr,
        "argstr": argstr,
        "str_sol": str_sol,
        "str_stu": str_stu,
        "action": action_strs[test],
    }

    # either error test and no error, or vice-versa
    stu_node = state.student_parts["node"]
    stu_state = state.to_child(highlight=stu_node)
    if (test == "error") ^ isinstance(eval_stu, Exception):
        stu_state.report(error_msg, fmt_kwargs)

    # incorrect result
    stu_state.do_test(
        EqualTest(
            eval_sol, eval_stu, FeedbackComponent(incorrect_msg, fmt_kwargs), func
        )
    )

    return state



from protowhat.utils_messaging import get_ord
from tcs_pythonwhat.tasks import (
    getResultInProcess,
    getOutputInProcess,
    getErrorInProcess,
    ReprFail,
    isDefinedInProcess,
    getOptionFromProcess,
    UndefinedValue,
)
from tcs_pythonwhat.Test import EqualTest, DefinedCollTest
from protowhat.Feedback import Feedback, FeedbackComponent
from protowhat.failure import InstructorError, debugger
from tcs_pythonwhat import utils
from functools import partial
import re
import copy
import ast

evalCalls = {
    "value": getResultInProcess,
    "output": getOutputInProcess,
    "error": getErrorInProcess,
}


def has_part(state, name, msg, fmt_kwargs=None, index=None):
    d = {
        "sol_part": state.solution_parts,
        "stu_part": state.student_parts,
        **fmt_kwargs,
    }

    def verify(part, index):
        if index is not None:
            if isinstance(index, list):
                for ind in index:
                    part = part[ind]
            else:
                part = part[index]
        if part is None:
            raise KeyError

    # TODO: instructor error if msg is not str
    # Check if it's there in the solution
    try:
        verify(state.solution_parts[name], index)
    except (KeyError, IndexError):
        with debugger(state):
            err_msg = "SCT fails on solution: {}".format(msg)
            state.report(err_msg, d)

    try:
        verify(state.student_parts[name], index)
    except (KeyError, IndexError):
        state.report(msg, d)

    return state


def has_equal_part(state, name, msg):
    d = {
        "stu_part": state.student_parts,
        "sol_part": state.solution_parts,
        "name": name,
    }

    state.do_test(
        EqualTest(d["stu_part"][name], d["sol_part"][name], FeedbackComponent(msg, d))
    )

    return state


# TODO: shouldn't have to hardcode message
def has_equal_part_len(state, name, unequal_msg):
    """Verify that a part that is zoomed in on has equal length.

    Typically used in the context of ``check_function_def()``

    Arguments:
        name (str): name of the part for which to check the length to the corresponding part in the solution.
        unequal_msg (str): Message in case the lengths do not match.
        state (State): state as passed by the SCT chain. Don't specify this explicitly.

    :Examples:

        Student and solution code::

            def shout(word):
                return word + '!!!'

        SCT that checks number of arguments::

            Ex().check_function_def('shout').has_equal_part_len('args', 'not enough args!')
    """
    d = dict(
        stu_len=len(state.student_parts[name]), sol_len=len(state.solution_parts[name])
    )

    if d["stu_len"] != d["sol_len"]:
        state.report(unequal_msg, d)

    return state


# Expression tests -----------------------------------------------------------


def has_equal_ast(state, incorrect_msg=None, code=None, exact=True, append=None):
    """Test whether abstract syntax trees match between the student and solution code.

    ``has_equal_ast()`` can be used in two ways:

    * As a robust version of ``has_code()``. By setting ``code``, you can look for the AST representation of ``code`` in the student's submission.
      But be aware that ``a`` and ``a = 1`` won't match, as reading and assigning are not the same in an AST.
      Use ``ast.dump(ast.parse(code))`` to see an AST representation of ``code``.
    * As an expression-based check when using more advanced SCT chain, e.g. to compare the equality of expressions to set function arguments.

    Args:
        incorrect_msg: message displayed when ASTs mismatch. When you specify ``code`` yourself, you have to specify this.
        code: optional code to use instead of the solution AST.
        exact: whether the representations must match exactly. If false, the solution AST
               only needs to be contained within the student AST (similar to using test student typed).
               Defaults to ``True``, unless the ``code`` argument has been specified.

    :Example:

        Student and Solution Code::

            dict(a = 'value').keys()

        SCT::

            # all pass
            Ex().has_equal_ast()
            Ex().has_equal_ast(code = "dict(a = 'value').keys()")
            Ex().has_equal_ast(code = "dict(a = 'value')", exact = False)

        Student and Solution Code::

            import numpy as np
            arr = np.array([1, 2, 3, 4, 5])
            np.mean(arr)

        SCT::

            # Check underlying value of arugment a of np.mean:
            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()

            # Only check AST equality of expression used to specify argument a:
            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()

    """
    if utils.v2_only():
        state.assert_is_not(["object_assignments"], "has_equal_ast", ["check_object"])
        state.assert_is_not(["function_calls"], "has_equal_ast", ["check_function"])

    if code and incorrect_msg is None:
        raise InstructorError.from_message(
            "Если вы вручную указываете код для сопоставления внутри has_equal_as(), "
            "вам необходимо явно задать аргумент `incorrect_msg`"
        )

    if (
        append is None
    ):  # if not specified, set to False if incorrect_msg was manually specified
        append = incorrect_msg is None
    if incorrect_msg is None:
        incorrect_msg = "Expected `{{sol_str}}`, but got `{{stu_str}}`."

    def parse_tree(tree):
        # get contents of module.body if only 1 element
        crnt = (
            tree.body[0]
            if isinstance(tree, ast.Module) and len(tree.body) == 1
            else tree
        )

        # remove Expr if it exists
        return ast.dump(crnt.value if isinstance(crnt, ast.Expr) else crnt)

    stu_rep = parse_tree(state.student_ast)
    sol_rep = parse_tree(state.solution_ast if not code else ast.parse(code))

    fmt_kwargs = {
        "sol_str": state.solution_code if not code else code,
        "stu_str": state.student_code,
    }

    if exact and not code:
        state.do_test(
            EqualTest(
                stu_rep,
                sol_rep,
                FeedbackComponent(incorrect_msg, fmt_kwargs, append=append),
            )
        )
    elif sol_rep not in stu_rep:
        state.report(incorrect_msg, fmt_kwargs, append=append)

    return state


DEFAULT_INCORRECT_MSG = "Ожидалось {{test_desc}}`{{sol_eval}}`, но было получено `{{stu_eval}}`."
DEFAULT_ERROR_MSG = "Выполнение {{'it' if parent['part'] else 'the highlighted expression'}} привело к ошибке: `{{stu_str}}`."
DEFAULT_ERROR_MSG_INV = "Выполнение {{'it' if parent['part'] else 'the highlighted expression'}} не привело к ошибке:, но оно должно было!"
DEFAULT_UNDEFINED_NAME_MSG = "Выполнение {{'it' if parent['part'] else 'the highlighted expression'}} должно определить переменную `{{name}}` без ошибок, но были получены ошибки."
DEFAULT_INCORRECT_NAME_MSG = (
    "Вы уверены, что присвоили правильное значение `{{name}}`?"
)
DEFAULT_INCORRECT_EXPR_CODE_MSG = (
    "Выполнение выражения `{{expr_code}}` не привело к ожидаемому результату."
)

args_string = """

    Args:
        incorrect_msg (str): feedback message if the {0} of the expression in the solution
          doesn't match the one of the student. This feedback message will be expanded if it is used
          in the context of another check function, like ``check_if_else``.
        error_msg (str): feedback message if there was an error when running the targeted student code.
          Note that when testing for an error, this message is displayed when none is raised.
        undefined_msg (str): feedback message if the ``name`` argument is defined, but a variable
          with that name doesn't exist after running the targeted student code.
        extra_env (dict): set variables to the extra environment. They will update the student and solution environment in
          the active state before the student/solution code in the active state is ran. This argument should contain a
          dictionary with the keys the names of the variables you want to set, and the values are the values of these variables.
          You can also use ``set_env()`` for this.
        context_vals (list): set variables which are bound in a ``for`` loop to certain values.
          This argument is only useful when checking a for loop (or list comprehensions).
          It contains a list with the values of the bound variables.
          You can also use ``set_context()`` for this.
        pre_code (str): the code in string form that should be executed before the expression is executed.
          This is the ideal place to set a random seed, for example.
        expr_code (str): If this argument is set, the expression in the student/solution code will not
          be ran. Instead, the given piece of code will be ran in the student as well as the solution environment
          and the result will be compared. However if the string contains one or more placeholders ``__focus__``,
          they will be substituted by the currently focused code.
        name (str): If this is specified, the {0} of running this expression after running the focused expression
          is returned, instead of the {0} of the focused expression in itself. This is typically used to inspect the
          {0} of an object after executing the body of e.g. a ``for`` loop.
        copy (bool): whether to try to deep copy objects in the environment, such as lists, that could
          accidentally be mutated. Disable to speed up SCTs. Disabling may lead to cryptic mutation issues.
        func (function): custom binary function of form f(stu_result, sol_result), for equality testing.
        override: If specified, this avoids the execution of the targeted code in the solution process. Instead, it
          will compare the {0} of the expression in the student process with the value specified in ``override``.
          Typically used in a ``SingleProcessExercise`` or if you want to allow for different solutions other than
          the one coded up in the solution.
    """


def has_expr(
    state,
    incorrect_msg=None,
    error_msg=None,
    undefined_msg=None,
    append=None,
    extra_env=None,
    context_vals=None,
    pre_code=None,
    expr_code=None,
    name=None,
    copy=True,
    func=None,
    override=None,
    test=None,  # todo: default or arg before state
):

    if (
        append is None
    ):  # if not specified, set to False if incorrect_msg was manually specified
        append = incorrect_msg is None
    if incorrect_msg is None:
        if name:
            incorrect_msg = DEFAULT_INCORRECT_NAME_MSG
        elif expr_code:
            incorrect_msg = DEFAULT_INCORRECT_EXPR_CODE_MSG
        else:
            incorrect_msg = DEFAULT_INCORRECT_MSG
    if undefined_msg is None:
        undefined_msg = DEFAULT_UNDEFINED_NAME_MSG
    if error_msg is None:
        if test == "error":
            error_msg = DEFAULT_ERROR_MSG_INV
        else:
            error_msg = DEFAULT_ERROR_MSG

    if state.solution_code is not None and isinstance(expr_code, str):
        expr_code = expr_code.replace("__focus__", state.solution_code)

    get_func = partial(
        evalCalls[test],
        extra_env=extra_env,
        context_vals=context_vals,
        pre_code=pre_code,
        expr_code=expr_code,
        name=name,
        copy=copy,
    )

    if override is not None:
        # don't bother with running expression and fetching output/value
        # eval_sol, str_sol = eval
        eval_sol, str_sol = override, str(override)
    else:
        eval_sol, str_sol = get_func(
            tree=state.solution_ast,
            process=state.solution_process,
            context=state.solution_context,
            env=state.solution_env,
        )

        if (test == "error") ^ isinstance(eval_sol, Exception):
            raise InstructorError.from_message(
                "Evaluating expression raised error in solution process (or didn't raise if testing for one). "
                "Error: {} - {}".format(type(eval_sol), str_sol)
            )
        if isinstance(eval_sol, ReprFail):
            raise InstructorError.from_message(
                "Couldn't extract the value for the highlighted expression from the solution process: "
                + eval_sol.info
            )

    eval_stu, str_stu = get_func(
        tree=state.student_ast,
        process=state.student_process,
        context=state.student_context,
        env=state.student_env,
    )

    # kwargs ---
    fmt_kwargs = {
        "stu_part": state.student_parts,
        "sol_part": state.solution_parts,
        "name": name,
        "test": test,
        "test_desc": "" if test == "value" else "the %s " % test,
        "expr_code": expr_code,
    }

    fmt_kwargs["stu_eval"] = str(eval_stu)
    fmt_kwargs["sol_eval"] = str(eval_sol)

    # wrap in quotes if eval_sol or eval_stu are strings
    if test == "value":
        if isinstance(eval_stu, str):
            fmt_kwargs["stu_eval"] = '\'{}\''.format(fmt_kwargs["stu_eval"])
        if isinstance(eval_sol, str):
            fmt_kwargs["sol_eval"] = '\'{}\''.format(fmt_kwargs["sol_eval"])

    # reformat student evaluation string if it is too long
    fmt_kwargs["stu_eval"] = utils.shorten_string(fmt_kwargs["stu_eval"])

    # check if student or solution evaluations are too long or contain newlines
    if incorrect_msg == DEFAULT_INCORRECT_MSG and (
        len(fmt_kwargs["sol_eval"]) > 50 or
        utils.has_newline(fmt_kwargs["stu_eval"]) or
        utils.has_newline(fmt_kwargs["sol_eval"]) or
        fmt_kwargs["stu_eval"] == fmt_kwargs["sol_eval"]):
        fmt_kwargs["stu_eval"] = None
        fmt_kwargs["sol_eval"] = None
        incorrect_msg = "Expected something different."

    # tests ---
    # error in process
    if (test == "error") ^ isinstance(eval_stu, Exception):
        fmt_kwargs["stu_str"] = str_stu
        state.report(error_msg, fmt_kwargs, append=append)

    # name is undefined after running expression
    if isinstance(eval_stu, UndefinedValue):
        state.report(undefined_msg, fmt_kwargs, append=append)

    # test equality of results
    state.do_test(
        EqualTest(
            eval_stu,
            eval_sol,
            FeedbackComponent(incorrect_msg, fmt_kwargs, append=append),
            func,
        )
    )

    return state


has_equal_value = partial(has_expr, test="value")
has_equal_value.__name__ = "has_equal_value"
has_equal_value.__doc__ = (
    """Run targeted student and solution code, and compare returned value.

    When called on an SCT chain, ``has_equal_value()`` will execute the student and solution
    code that is 'zoomed in on' and compare the returned values.
    """
    + args_string.format("returned value", "value")
    + """
    :Example:

        Student code and solution code::

            import numpy as np
            arr = np.array([1, 2, 3, 4, 5])
            np.mean(arr)

        SCT::

            # Verify equality of arr:
            Ex().check_object('arr').has_equal_value()

            # Verify whether arr was correctly set in np.mean
            Ex().check_function('numpy.mean').check_args('a').has_equal_value()

            # Verify whether np.mean(arr) produced the same result
            Ex().check_function('numpy.mean').has_equal_value()

    """
)


has_equal_output = partial(has_expr, test="output")
has_equal_output.__name__ = "has_equal_output"
has_equal_output.__doc__ = """Run targeted student and solution code, and compare output.

    When called on an SCT chain, ``has_equal_output()`` will execute the student and solution
    code that is 'zoomed in on' and compare the output.
    """ + args_string.format(
    "output"
)

has_equal_error = partial(has_expr, test="error")
has_equal_error.__name__ = "has_equal_error"
has_equal_error.__doc__ = """Run targeted student and solution code, and compare generated errors.

    When called on an SCT chain, ``has_equal_error()`` will execute the student and solution
    code that is 'zoomed in on' and compare the errors that they generate.
    """ + args_string.format(
    "error"
)

## Various has tests ----------------------------------------------------------

from tcs_pythonwhat.Test import StringContainsTest


def has_code(state, text, pattern=True, not_typed_msg=None):
    """Test the student code.

    Tests if the student typed a (pattern of) text. It is advised to use ``has_equal_ast()`` instead of ``has_code()``,
    as it is more robust to small syntactical differences that don't change the code's behavior.

    Args:
        text (str): the text that is searched for
        pattern (bool): if True (the default), the text is treated as a pattern. If False, it is treated as plain text.
        not_typed_msg (str): feedback message to be displayed if the student did not type the text.

    :Example:

        Student code and solution code::

            y = 1 + 2 + 3

        SCT::

            # Verify that student code contains pattern (not robust!!):
            Ex().has_code(r"1\\s*\\+2\\s*\\+3")

    """
    if not not_typed_msg:
        if pattern:
            not_typed_msg = "Не удалось найти правильный шаблон в вашем коде."
        else:
            not_typed_msg = "Не удалось найти следующий текст в вашем коде: %r" % text

    student_code = state.student_code

    state.do_test(StringContainsTest(student_code, text, pattern, not_typed_msg))

    return state


def has_import(
    state,
    name,
    same_as=False,
    not_imported_msg="Вы импортировали `{{pkg}}`?",
    incorrect_as_msg="Вы импортировали `{{pkg}}` как `{{alias}}`?",
):
    """Checks whether student imported a package or function correctly.

    Python features many ways to import packages.
    All of these different methods revolve around the ``import``, ``from`` and ``as`` keywords.
    ``has_import()`` provides a robust way to check whether a student correctly imported a certain package.

    By default, ``has_import()`` allows for different ways of aliasing the imported package or function.
    If you want to make sure the correct alias was used to refer to the package or function that was imported,
    set ``same_as=True``.

    Args:
        name (str): the name of the package that has to be checked.
        same_as (bool): if True, the alias of the package or function has to be the same. Defaults to False.
        not_imported_msg (str): feedback message when the package is not imported.
        incorrect_as_msg (str): feedback message if the alias is wrong.

    :Example:

        Example 1, where aliases don't matter (defaut): ::

            # solution
            import matplotlib.pyplot as plt

            # sct
            Ex().has_import("matplotlib.pyplot")

            # passing submissions
            import matplotlib.pyplot as plt
            from matplotlib import pyplot as plt
            import matplotlib.pyplot as pltttt

            # failing submissions
            import matplotlib as mpl

        Example 2, where the SCT is coded so aliases do matter: ::

            # solution
            import matplotlib.pyplot as plt

            # sct
            Ex().has_import("matplotlib.pyplot", same_as=True)

            # passing submissions
            import matplotlib.pyplot as plt
            from matplotlib import pyplot as plt

            # failing submissions
            import matplotlib.pyplot as pltttt

    """
    student_imports = state.ast_dispatcher.find("imports", state.student_ast)
    solution_imports = state.ast_dispatcher.find("imports", state.solution_ast)

    if name not in solution_imports:
        raise InstructorError.from_message(
            "`has_import()` не смог найти импортирование библиотеки %s в solution-коде."
            % name
        )

    fmt_kwargs = {"pkg": name, "alias": solution_imports[name]}

    state.do_test(
        DefinedCollTest(
            name, student_imports, FeedbackComponent(not_imported_msg, fmt_kwargs)
        )
    )

    if same_as:
        state.do_test(
            EqualTest(
                solution_imports[name],
                student_imports[name],
                FeedbackComponent(incorrect_as_msg, fmt_kwargs),
            )
        )

    return state


def has_output(state, text, pattern=True, no_output_msg=None):
    """Search student output for a pattern.

    Among the student and solution process, the student submission and solution code as a string,
    the ``Ex()`` state also contains the output that a student generated with his or her submission.

    With ``has_output()``, you can access this output and match it against a regular or fixed expression.

    Args:
        text (str): the text that is searched for
        pattern (bool): if True (default), the text is treated as a pattern. If False, it is treated as plain text.
        no_output_msg (str): feedback message to be displayed if the output is not found.

    :Example:

        As an example, suppose we want a student to print out a sentence: ::

            # Print the "This is some ... stuff"
            print("This is some weird stuff")

        The following SCT tests whether the student prints out ``This is some weird stuff``: ::

            # Using exact string matching
            Ex().has_output("This is some weird stuff", pattern = False)

            # Using a regular expression (more robust)
            # pattern = True is the default
            msg = "Print out ``This is some ... stuff`` to the output, " + \\
                  "fill in ``...`` with a word you like."
            Ex().has_output(r"This is some \w* stuff", no_output_msg = msg)

    """
    if not no_output_msg:
        no_output_msg = "Вы не вывели ожидаемый результат."

    state.do_test(
        StringContainsTest(state.raw_student_output, text, pattern, no_output_msg)
    )

    return state


def has_printout(
    state, index, not_printed_msg=None, pre_code=None, name=None, copy=False
):
    """Check if the right printouts happened.

    ``has_printout()`` will look for the printout in the solution code that you specified with ``index`` (0 in this case), rerun the ``print()`` call in
    the solution process, capture its output, and verify whether the output is present in the output of the student.

    This is more robust as ``Ex().check_function('print')`` initiated chains as students can use as many
    printouts as they want, as long as they do the correct one somewhere.

    Args:
        index (int): index of the ``print()`` call in the solution whose output you want to search for in the student output.
        not_printed_msg (str): if specified, this overrides the default message that is generated when the output
          is not found in the student output.
        pre_code (str): Python code as a string that is executed before running the targeted student call.
          This is the ideal place to set a random seed, for example.
        copy (bool): whether to try to deep copy objects in the environment, such as lists, that could
          accidentally be mutated. Disabled by default, which speeds up SCTs.
        state (State): state as passed by the SCT chain. Don't specify this explicitly.

    :Example:

        Suppose you want somebody to print out 4: ::

            print(1, 2, 3, 4)

        The following SCT would check that: ::

            Ex().has_printout(0)

        All of the following SCTs would pass: ::

            print(1, 2, 3, 4)
            print('1 2 3 4')
            print(1, 2, '3 4')
            print("random"); print(1, 2, 3, 4)

    :Example:

        Watch out: ``has_printout()`` will effectively **rerun** the ``print()`` call in the solution process after the entire solution script was executed.
        If your solution script updates the value of `x` after executing it, ``has_printout()`` will not work.

        Suppose you have the following solution: ::

            x = 4
            print(x)
            x = 6

        The following SCT will not work: ::

            Ex().has_printout(0)

        Why? When the ``print(x)`` call is executed, the value of ``x`` will be 6, and pythonwhat will look for the output `'6`' in the output the student generated.
        In cases like these, ``has_printout()`` cannot be used.

    :Example:

        Inside a for loop ``has_printout()``

        Suppose you have the following solution: ::

            for i in range(5):
                print(i)

        The following SCT will not work: ::

            Ex().check_for_loop().check_body().has_printout(0)

        The reason is that ``has_printout()`` can only be called from the root state. ``Ex()``.
        If you want to check printouts done in e.g. a for loop, you have to use a `check_function('print')` chain instead: ::

            Ex().check_for_loop().check_body().\\
                set_context(0).check_function('print').\\
                check_args(0).has_equal_value()

    """

    extra_msg = "If you want to check printouts done in e.g. a for loop, you have to use a `check_function('print')` chain instead."
    state.assert_execution_root("has_printout", extra_msg=extra_msg)

    if not_printed_msg is None:
        not_printed_msg = (
            "Вы использовали `{{sol_call}}` чтобы сделать вывод соответствующего результата?"
        )

    try:
        sol_call_ast = state.ast_dispatcher.find("function_calls", state.solution_ast)[
            "print"
        ][index]["node"]
    except (KeyError, IndexError):
        raise InstructorError.from_message(
            "`has_printout({})` не смог найти вызов {} для вывода соответствующего результата в solution-коде.".format(
                index, get_ord(index + 1)
            )
        )

    out_sol, str_sol = getOutputInProcess(
        tree=sol_call_ast,
        process=state.solution_process,
        context=state.solution_context,
        env=state.solution_env,
        pre_code=pre_code,
        copy=copy,
    )

    sol_call_str = state.solution_ast_tokens.get_text(sol_call_ast)

    if isinstance(str_sol, Exception):
        with debugger(state):
            state.report(
                "Evaluating the solution expression {} raised error in solution process."
                "Error: {} - {}".format(sol_call_str, type(out_sol), str_sol)
            )

    has_output(
        state,
        out_sol.strip(),
        pattern=False,
        no_output_msg=FeedbackComponent(not_printed_msg, {"sol_call": sol_call_str}),
    )

    return state


def has_no_error(
    state,
    incorrect_msg="Взгляните на консоль: ваш код содержит ошибку. Исправьте это и попробуйте еще раз!",
):
    """Check whether the submission did not generate a runtime error.

    If all SCTs for an exercise pass, before marking the submission as correct pythonwhat will automatically check whether
    the student submission generated an error. This means it is not needed to use ``has_no_error()`` explicitly.

    However, in some cases, using ``has_no_error()`` explicitly somewhere throughout your SCT execution can be helpful:

    - If you want to make sure people didn't write typos when writing a long function name.
    - If you want to first verify whether a function actually runs, before checking whether the arguments were specified correctly.
    - More generally, if, because of the content, it's instrumental that the script runs without
      errors before doing any other verifications.

    Args:
        incorrect_msg: if specified, this overrides the default message if the student code generated an error.

    :Example:

        Suppose you're verifying an exercise about model training and validation: ::

            # pre exercise code
            import numpy as np
            from sklearn.model_selection import train_test_split
            from sklearn import datasets
            from sklearn import svm

            iris = datasets.load_iris()
            iris.data.shape, iris.target.shape

            # solution
            X_train, X_test, y_train, y_test = train_test_split(
                iris.data, iris.target, test_size=0.4, random_state=0)

        If you want to make sure that ``train_test_split()`` ran without errors,
        which would check if the student typed the function without typos and used
        sensical arguments, you could use the following SCT: ::

            Ex().has_no_error()
            Ex().check_function('sklearn.model_selection.train_test_split').multi(
                check_args(['arrays', 0]).has_equal_value(),
                check_args(['arrays', 0]).has_equal_value(),
                check_args(['options', 'test_size']).has_equal_value(),
                check_args(['options', 'random_state']).has_equal_value()
            )

        If, on the other hand, you want to fall back onto pythonwhat's built in behavior,
        that checks for an error before marking the exercise as correct, you can simply
        leave of the ``has_no_error()`` step.

    """
    state.assert_execution_root("has_no_error")

    if state.reporter.errors:
        state.report(incorrect_msg, {"error": str(state.reporter.errors[0])})

    return state


MC_VAR_NAME = "selected_option"


def has_chosen(state, correct, msgs):
    """Test multiple choice exercise.

    Test for a MultipleChoiceExercise. The correct answer (as an integer) and feedback messages
    are passed to this function.

    Args:
        correct (int): the index of the correct answer (should be an instruction). Starts at 1.
        msgs (list(str)): a list containing all feedback messages belonging to each choice of the
                          student. The list should have the same length as the number of options.
    """
    if not issubclass(type(correct), int):
        raise InstructorError.from_message(
            "Внутри `has_chosen()` аргумент 'correct' должен быть целым числом."
        )

    student_process = state.student_process
    if not isDefinedInProcess(MC_VAR_NAME, student_process):
        raise InstructorError.from_message(
            "Опция не доступна в student-коде"
        )
    else:
        selected_option = getOptionFromProcess(student_process, MC_VAR_NAME)
        if not issubclass(type(selected_option), int):
            raise InstructorError.from_message("selected_option должен быть целым числом")

        if selected_option < 1 or correct < 1:
            raise InstructorError.from_message(
                "selected_option and correct должны быть больше нуля"
            )

        if selected_option > len(msgs) or correct > len(msgs):
            raise InstructorError.from_message(
                "определено недостаточно сообщений обратной связи."
            )

        feedback_msg = msgs[selected_option - 1]

        state.reporter.success_msg = msgs[correct - 1]

        state.do_test(EqualTest(selected_option, correct, feedback_msg))



import ast


def assign_from_ast(node, expr):
    """
    Creates code to assign name (or tuple of names) node from expr

    This is useful for recreating destructuring assignment behavior, like
    a, *b = [1,2,3].
    """
    if isinstance(expr, str):
        expr = ast.Name(id=expr, ctx=ast.Load())
    mod = ast.Module([ast.Assign(targets=[node], value=expr)], [])
    ast.fix_missing_locations(mod)
    return compile(mod, "<assignment_script>", "exec")


def set_context_vals(env, context, context_vals=None):
    env.update(**context)
    # support old test_* functions with context_val arguments
    if context_vals:
        crnt_ctx = context.context
        if len(crnt_ctx) == 1 != len(context_vals):
            context_vals = [context_vals]

        env.update({key: value for (key, value) in zip(crnt_ctx, context_vals)})



from tcs_pythonwhat.Feedback import FeedbackComponent
from tcs_pythonwhat.checks.check_funcs import part_to_child
from tcs_pythonwhat.tasks import getSignatureInProcess
from protowhat.utils_messaging import get_ord, get_times
from protowhat.failure import debugger
from tcs_pythonwhat.parsing import IndexedDict
from functools import partial


def bind_args(signature, args_part):
    pos_args = []
    kw_args = {}
    for k, arg in args_part.items():
        if isinstance(k, int):
            pos_args.append(arg)
        else:
            kw_args[k] = arg

    bound_args = signature.bind(*pos_args, **kw_args)
    return IndexedDict(bound_args.arguments)


def get_mapped_name(name, mappings):
    # get name by splitting on periods
    if "." in name:
        for orig, full_name in mappings.items():
            if name.startswith(full_name):
                return name.replace(full_name, orig)
    return name


MISSING_MSG = "Вы вызвали функцию `{{mapped_name}}()`{{' ' + times if index>0}}?"
SIG_ISSUE_MSG = (
    "Вы указали аргументы для функции `{{mapped_name}}()`, используя корректный синтаксис?"
)
PREPEND_MSG = "Проверьте {{ord + ' ' if index>0}} вызов `{{mapped_name}}()`. "


def check_function(
    state,
    name,
    index=0,
    missing_msg=None,
    params_not_matched_msg=None,
    expand_msg=None,
    signature=True,
):
    """Check whether a particular function is called.

    ``check_function()`` is typically followed by:

    - ``check_args()`` to check whether the arguments were specified.
      In turn, ``check_args()`` can be followed by ``has_equal_value()`` or ``has_equal_ast()``
      to assert that the arguments were correctly specified.
    - ``has_equal_value()`` to check whether rerunning the function call coded by the student
      gives the same result as calling the function call as in the solution.

    Checking function calls is a tricky topic. Please visit the
    `dedicated article <articles/checking_function_calls.html>`_ for more explanation,
    edge cases and best practices.

    Args:
        name (str): the name of the function to be tested. When checking functions in packages, always
            use the 'full path' of the function.
        index (int): index of the function call to be checked. Defaults to 0.
        missing_msg (str): If specified, this overrides an automatically generated feedback message in case
            the student did not call the function correctly.
        params_not_matched_msg (str): If specified, this overrides an automatically generated feedback message
            in case the function parameters were not successfully matched.
        expand_msg (str): If specified, this overrides any messages that are prepended by previous SCT chains.
        signature (Signature): Normally, check_function() can figure out what the function signature is,
            but it might be necessary to use ``sig_from_params()`` to manually build a signature and pass this along.
        state (State): State object that is passed from the SCT Chain (don't specify this).

    :Examples:

        Student code and solution code::

            import numpy as np
            arr = np.array([1, 2, 3, 4, 5])
            np.mean(arr)

        SCT::

            # Verify whether arr was correctly set in np.mean
            Ex().check_function('numpy.mean').check_args('a').has_equal_value()

            # Verify whether np.mean(arr) produced the same result
            Ex().check_function('numpy.mean').has_equal_value()
    """

    append_missing = missing_msg is None
    append_params_not_matched = params_not_matched_msg is None
    if missing_msg is None:
        missing_msg = MISSING_MSG
    if expand_msg is None:
        expand_msg = PREPEND_MSG
    if params_not_matched_msg is None:
        params_not_matched_msg = SIG_ISSUE_MSG

    stu_out = state.ast_dispatcher.find("function_calls", state.student_ast)
    sol_out = state.ast_dispatcher.find("function_calls", state.solution_ast)

    student_mappings = state.ast_dispatcher.find("mappings", state.student_ast)

    fmt_kwargs = {
        "times": get_times(index + 1),
        "ord": get_ord(index + 1),
        "index": index,
        "mapped_name": get_mapped_name(name, student_mappings),
    }

    # Get Parts ----
    # Copy, otherwise signature binding overwrites sol_out[name][index]['args']
    with debugger(state):
        try:
            sol_parts = {**sol_out[name][index]}
        except KeyError:
            state.report(
                "`check_function()` не удается найти вызов `%s()` в solution-коде. Убедитесь, что вы правильно составили ожидания для кода!"
                % name
            )
        except IndexError:
            state.report(
                "`check_function()` не удается найти %s вызовы `%s()` в solution-коде."
                % (index + 1, name)
            )

    try:
        # Copy, otherwise signature binding overwrites stu_out[name][index]['args']
        stu_parts = {**stu_out[name][index]}
    except (KeyError, IndexError):
        state.report(missing_msg, fmt_kwargs, append=append_missing)

    # Signatures -----
    if signature:
        signature = None if isinstance(signature, bool) else signature
        get_sig = partial(
            getSignatureInProcess,
            name=name,
            signature=signature,
            manual_sigs=state.get_manual_sigs(),
        )

        try:
            sol_sig = get_sig(
                mapped_name=sol_parts["name"], process=state.solution_process
            )
            sol_parts["args"] = bind_args(sol_sig, sol_parts["args"])
        except Exception as e:
            with debugger(state):
                state.report(
                    "`check_function()` не удается сопоставить %s вызовы `%s` их сигнатурам:\n%s "
                    % (get_ord(index + 1), name, e)
                )

        try:
            stu_sig = get_sig(
                mapped_name=stu_parts["name"], process=state.student_process
            )
            stu_parts["args"] = bind_args(stu_sig, stu_parts["args"])
        except Exception:
            state.to_child(highlight=stu_parts["node"]).report(
                params_not_matched_msg, fmt_kwargs, append=append_params_not_matched
            )

    # three types of parts: pos_args, keywords, args (e.g. these are bound to sig)
    append_message = FeedbackComponent(expand_msg, fmt_kwargs)
    child = part_to_child(
        stu_parts, sol_parts, append_message, state, node_name="function_calls"
    )
    return child



from tcs_pythonwhat.parsing import ObjectAssignmentParser
from tcs_pythonwhat.Test import (
    DefinedProcessTest,
    InstanceProcessTest,
    DefinedCollProcessTest,
)
from protowhat.Feedback import FeedbackComponent
from protowhat.failure import InstructorError
from tcs_pythonwhat.tasks import (
    isDefinedInProcess,
    isInstanceInProcess,
    isDefinedCollInProcess,
)
from tcs_pythonwhat.checks.check_funcs import part_to_child
from tcs_pythonwhat.utils import v2_only
import pandas as pd
import ast


def check_object(state, index, missing_msg=None, expand_msg=None, typestr="variable"):
    """Check object existence (and equality)

    Check whether an object is defined in the student's process, and zoom in on its value in both
    student and solution process to inspect quality (with has_equal_value().

    In ``pythonbackend``, both the student's submission as well as the solution code are executed, in separate processes.
    ``check_object()`` looks at these processes and checks if the referenced object is available in the student process.
    Next, you can use ``has_equal_value()`` to check whether the objects in the student and solution process correspond.

    Args:
        index (str): the name of the object which value has to be checked.
        missing_msg (str): feedback message when the object is not defined in the student process.
        expand_msg (str): If specified, this overrides any messages that are prepended by previous SCT chains.

    :Example:

        Suppose you want the student to create a variable ``x``, equal to 15: ::

            x = 15

        The following SCT will verify this: ::

            Ex().check_object("x").has_equal_value()

        - ``check_object()`` will check if the variable ``x`` is defined in the student process.
        - ``has_equal_value()`` will check whether the value of ``x`` in the solution process is the same as in the student process.

        Note that ``has_equal_value()`` only looks at **end result** of a variable in the student process.
        In the example, how the object ``x`` came about in the student's submission, does not matter.
        This means that all of the following submission will also pass the above SCT: ::

            x = 15
            x = 12 + 3
            x = 3; x += 12

    :Example:

        As the previous example mentioned, ``has_equal_value()`` only looks at the **end result**. If your exercise is
        first initializing and object and further down the script is updating the object, you can only look at the final value!

        Suppose you want the student to initialize and populate a list `my_list` as follows: ::

            my_list = []
            for i in range(20):
                if i % 3 == 0:
                    my_list.append(i)

        There is no robust way to verify whether `my_list = [0]` was coded correctly in a separate way.
        The best SCT would look something like this: ::

            msg = "Have you correctly initialized `my_list`?"
            Ex().check_correct(
                check_object('my_list').has_equal_value(),
                multi(
                    # check initialization: [] or list()
                    check_or(
                        has_equal_ast(code = "[]", incorrect_msg = msg),
                        check_function('list')
                    ),
                    check_for_loop().multi(
                        check_iter().has_equal_value(),
                        check_body().check_if_else().multi(
                            check_test().multi(
                                set_context(2).has_equal_value(),
                                set_context(3).has_equal_value()
                            ),
                            check_body().set_context(3).\\
                                set_env(my_list = [0]).\\
                                has_equal_value(name = 'my_list')
                        )
                    )
                )
            )

        - ``check_correct()`` is used to robustly check whether ``my_list`` was built correctly.
        - If ``my_list`` is not correct, **both** the initialization and the population code are checked.

    :Example:

        Because checking object correctness incorrectly is such a common misconception, we're adding another example: ::

            import pandas as pd
            df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
            df['c'] = [7, 8, 9]

        The following SCT would be **wrong**, as it does not factor in the possibility that the 'add column ``c``' step could've been wrong: ::

            Ex().check_correct(
                check_object('df').has_equal_value(),
                check_function('pandas.DataFrame').check_args(0).has_equal_value()
            )

        The following SCT would be better, as it is specific to the steps: ::

            # verify the df = pd.DataFrame(...) step
            Ex().check_correct(
                check_df('df').multi(
                    check_keys('a').has_equal_value(),
                    check_keys('b').has_equal_value()
                ),
                check_function('pandas.DataFrame').check_args(0).has_equal_value()
            )

            # verify the df['c'] = [...] step
            Ex().check_df('df').check_keys('c').has_equal_value()

    :Example:

        pythonwhat compares the objects in the student and solution process with the ``==`` operator.
        For basic objects, this ``==`` is operator is properly implemented, so that the objects can be effectively compared.
        For more complex objects that are produced by third-party packages, however, it's possible that this equality operator is not implemented in a way you'd expect.
        Often, for these object types the ``==`` will compare the actual object instances: ::

            # pre exercise code
            class Number():
                def __init__(self, n):
                    self.n = n

            # solution
            x = Number(1)

            # sct that won't work
            Ex().check_object().has_equal_value()

            # sct
            Ex().check_object().has_equal_value(expr_code = 'x.n')

            # submissions that will pass this sct
            x = Number(1)
            x = Number(2 - 1)

        The basic SCT like in the previous example will notwork here.
        Notice how we used the ``expr_code`` argument to _override_ which value `has_equal_value()` is checking.
        Instead of checking whether `x` corresponds between student and solution process, it's now executing the expression ``x.n``
        and seeing if the result of running this expression in both student and solution process match.

    """

    # Only do the assertion if PYTHONWHAT_V2_ONLY is set to '1'
    if v2_only():
        extra_msg = "Если вы хотите проверить значение объекта, например, в цикле for, используйте `has_equal_value(имя = 'my_obj')`."
        state.assert_execution_root("check_object", extra_msg=extra_msg)

    if missing_msg is None:
        missing_msg = "Вы определили {{typestr}} `{{index}}` без ошибок?"

    if expand_msg is None:
        expand_msg = "Вы правильно определили {{typestr}} `{{index}}`? "

    if (
        not isDefinedInProcess(index, state.solution_process)
        and state.has_different_processes()
    ):
        raise InstructorError.from_message(
            "`check_object()` не может найти объект `%s` в solution-коде."
            % index
        )

    append_message = FeedbackComponent(expand_msg, {"index": index, "typestr": typestr})

    # create child state, using either parser output, or create part from name
    fallback = lambda: ObjectAssignmentParser.get_part(index)
    stu_part = state.ast_dispatcher.find("object_assignments", state.student_ast).get(
        index, fallback()
    )
    sol_part = state.ast_dispatcher.find("object_assignments", state.solution_ast).get(
        index, fallback()
    )

    # test object exists
    state.do_test(
        DefinedProcessTest(
            index,
            state.student_process,
            FeedbackComponent(missing_msg, append_message.kwargs),
        )
    )

    child = part_to_child(
        stu_part, sol_part, append_message, state, node_name="object_assignments"
    )

    return child


def is_instance(state, inst, not_instance_msg=None):
    """Check whether an object is an instance of a certain class.

    ``is_instance()`` can currently only be used when chained from ``check_object()``, the function that is
    used to 'zoom in' on the object of interest.

    Args:
        inst (class): The class that the object should have.
        not_instance_msg (str): When specified, this overrides the automatically generated message in case
            the object does not have the expected class.
        state (State): The state that is passed in through the SCT chain (don't specify this).

    :Example:

        Student code and solution code::

            import numpy as np
            arr = np.array([1, 2, 3, 4, 5])

        SCT::

            # Verify the class of arr
            import numpy
            Ex().check_object('arr').is_instance(numpy.ndarray)
    """

    state.assert_is(["object_assignments"], "is_instance", ["check_object"])

    sol_name = state.solution_parts.get("name")
    stu_name = state.student_parts.get("name")

    if not_instance_msg is None:
        not_instance_msg = "Is it a {{inst.__name__}}?"

    if not isInstanceInProcess(sol_name, inst, state.solution_process):
        raise InstructorError.from_message(
            "`is_instance()` выявило, что `%s` не является `%s` в solution-коде."
            % (sol_name, inst.__name__)
        )

    feedback = FeedbackComponent(not_instance_msg, {"inst": inst})
    state.do_test(InstanceProcessTest(stu_name, inst, state.student_process, feedback))

    return state


def check_df(state, index, missing_msg=None, not_instance_msg=None, expand_msg=None):
    """Check whether a DataFrame was defined and it is the right type

    ``check_df()`` is a combo of ``check_object()`` and ``is_instance()`` that checks whether the specified object exists
    and whether the specified object is pandas DataFrame.

    You can continue checking the data frame with ``check_keys()`` function to 'zoom in' on a particular column in the pandas DataFrame:

    Args:
        index (str): Name of the data frame to zoom in on.
        missing_msg (str): See ``check_object()``.
        not_instance_msg (str): See ``is_instance()``.
        expand_msg (str): If specified, this overrides any messages that are prepended by previous SCT chains.

    :Example:

        Suppose you want the student to create a DataFrame ``my_df`` with two columns.
        The column ``a`` should contain the numbers 1 to 3,
        while the contents of column ``b`` can be anything: ::

            import pandas as pd
            my_df = pd.DataFrame({"a": [1, 2, 3], "b": ["a", "n", "y"]})

        The following SCT would robustly check that: ::

            Ex().check_df("my_df").multi(
                check_keys("a").has_equal_value(),
                check_keys("b")
            )

        - ``check_df()`` checks if ``my_df`` exists (``check_object()`` behind the scenes) and is a DataFrame (``is_instance()``)
        - ``check_keys("a")`` zooms in on the column ``a`` of the data frame, and ``has_equal_value()`` checks if the columns correspond between student and solution process.
        - ``check_keys("b")`` zooms in on hte column ``b`` of the data frame, but there's no 'equality checking' happening

        The following submissions would pass the SCT above: ::

            my_df = pd.DataFrame({"a": [1, 1 + 1, 3], "b": ["a", "l", "l"]})
            my_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})

    """
    child = check_object(
        state,
        index,
        missing_msg=missing_msg,
        expand_msg=expand_msg,
        typestr="pandas DataFrame",
    )
    is_instance(child, pd.DataFrame, not_instance_msg=not_instance_msg)
    return child


def check_keys(state, key, missing_msg=None, expand_msg=None):
    """Check whether an object (dict, DataFrame, etc) has a key.

    ``check_keys()`` can currently only be used when chained from ``check_object()``, the function that is
    used to 'zoom in' on the object of interest.

    Args:
        key (str): Name of the key that the object should have.
        missing_msg (str): When specified, this overrides the automatically generated
            message in case the key does not exist.
        expand_msg (str): If specified, this overrides any messages that are prepended by previous SCT chains.
        state (State): The state that is passed in through the SCT chain (don't specify this).

    :Example:

        Student code and solution code::

            x = {'a': 2}

        SCT::

            # Verify that x contains a key a
            Ex().check_object('x').check_keys('a')

            # Verify that x contains a key a and a is correct.
            Ex().check_object('x').check_keys('a').has_equal_value()

    """

    state.assert_is(["object_assignments"], "is_instance", ["check_object", "check_df"])

    if missing_msg is None:
        missing_msg = "Отсутствует {{ 'column' if 'DataFrame' in parent.typestr else 'key' }} `'{{key}}'`."
    if expand_msg is None:
        expand_msg = "Вы правильно указали {{ 'column' if 'DataFrame' in parent.typestr else 'key' }} `'{{key}}'`? "

    sol_name = state.solution_parts.get("name")
    stu_name = state.student_parts.get("name")

    if not isDefinedCollInProcess(sol_name, key, state.solution_process):
        raise InstructorError.from_message(
            "`check_keys()` не удалось найти ключ `%s` в объекте `%s` в solution-коде."
            % (key, sol_name)
        )

    # check if key available
    state.do_test(
        DefinedCollProcessTest(
            stu_name, key, state.student_process, FeedbackComponent(missing_msg, {"key": key})
        )
    )

    def get_part(name, key, highlight):
        if isinstance(key, str):
            slice_val = ast.Str(s=key)
        else:
            slice_val = ast.parse(str(key)).body[0].value
        expr = ast.Subscript(
            value=ast.Name(id=name, ctx=ast.Load()),
            slice=ast.Index(value=slice_val),
            ctx=ast.Load(),
        )
        ast.fix_missing_locations(expr)
        return {"node": expr, "highlight": highlight}

    stu_part = get_part(stu_name, key, state.student_parts.get("highlight"))
    sol_part = get_part(sol_name, key, state.solution_parts.get("highlight"))
    append_message = FeedbackComponent(expand_msg, {"key": key})
    child = part_to_child(stu_part, sol_part, append_message, state)
    return child



from protowhat.Feedback import FeedbackComponent
from tcs_pythonwhat.checks.check_logic import multi
from tcs_pythonwhat.checks.has_funcs import has_part
from protowhat.failure import debugger
from tcs_pythonwhat.tasks import setUpNewEnvInProcess, breakDownNewEnvInProcess
from protowhat.utils_messaging import get_ord
from tcs_pythonwhat.utils_ast import assert_ast
import ast
from jinja2 import Template


def render(template, kwargs):
    return Template(template).render(**kwargs)


def part_to_child(stu_part, sol_part, append_message, state, node_name=None):
    # stu_part and sol_part will be accessible on all templates
    append_message.kwargs.update({"stu_part": stu_part, "sol_part": sol_part})

    # if the parts are dictionaries, use to deck out child state
    if all(isinstance(p, dict) for p in [stu_part, sol_part]):
        child_state = state.to_child(
            student_ast=stu_part["node"],
            solution_ast=sol_part["node"],
            student_context=stu_part.get("target_vars"),
            solution_context=sol_part.get("target_vars"),
            student_parts=stu_part,
            solution_parts=sol_part,
            highlight=stu_part.get("highlight"),
            append_message=append_message,
            node_name=node_name,
        )
    else:
        # otherwise, assume they are just nodes
        child_state = state.to_child(
            student_ast=stu_part,
            solution_ast=sol_part,
            append_message=append_message,
            node_name=node_name,
        )

    return child_state


def check_part(state, name, part_msg, missing_msg=None, expand_msg=None):
    """Return child state with name part as its ast tree"""

    if missing_msg is None:
        missing_msg = "Вы уверены, что определили {{part}}? "
    if expand_msg is None:
        expand_msg = "Правильно ли вы указали {{part}}? "

    if not part_msg:
        part_msg = name
    append_message = FeedbackComponent(expand_msg, {"part": part_msg})

    has_part(state, name, missing_msg, append_message.kwargs)

    stu_part = state.student_parts[name]
    sol_part = state.solution_parts[name]

    assert_ast(state, sol_part, append_message.kwargs)

    return part_to_child(stu_part, sol_part, append_message, state)


def check_part_index(state, name, index, part_msg, missing_msg=None, expand_msg=None):
    """Return child state with indexed name part as its ast tree.

    ``index`` can be:

    - an integer, in which case the student/solution_parts are indexed by position.
    - a string, in which case the student/solution_parts are expected to be a dictionary.
    - a list of indices (which can be integer or string), in which case the student parts are indexed step by step.
    """

    if missing_msg is None:
        missing_msg = "Вы уверены, что определили {{part}}? "
    if expand_msg is None:
        expand_msg = "Правильно ли вы указали {{part}}? "

    # create message
    ordinal = get_ord(index + 1) if isinstance(index, int) else ""
    fmt_kwargs = {"index": index, "ordinal": ordinal}
    fmt_kwargs.update(part=render(part_msg, fmt_kwargs))

    append_message = FeedbackComponent(expand_msg, fmt_kwargs)

    # check there are enough parts for index
    has_part(state, name, missing_msg, fmt_kwargs, index)

    # get part at index
    stu_part = state.student_parts[name]
    sol_part = state.solution_parts[name]

    if isinstance(index, list):
        for ind in index:
            stu_part = stu_part[ind]
            sol_part = sol_part[ind]
    else:
        stu_part = stu_part[index]
        sol_part = sol_part[index]

    assert_ast(state, sol_part, fmt_kwargs)

    # return child state from part
    return part_to_child(stu_part, sol_part, append_message, state)


def check_node(
    state, name, index=0, typestr="{{ordinal}} node", missing_msg=None, expand_msg=None
):

    if missing_msg is None:
        missing_msg = "Система хочет проверить {{typestr}}, но он не был найден."
    if expand_msg is None:
        expand_msg = "Проверьте {{typestr}}. "

    stu_out = state.ast_dispatcher.find(name, state.student_ast)
    sol_out = state.ast_dispatcher.find(name, state.solution_ast)

    # check if there are enough nodes for index
    fmt_kwargs = {
        "ordinal": get_ord(index + 1) if isinstance(index, int) else "",
        "index": index,
        "name": name,
    }
    fmt_kwargs["typestr"] = render(typestr, fmt_kwargs)

    # test if node can be indexed succesfully
    try:
        stu_out[index]
    except (KeyError, IndexError):  # TODO comment errors
        state.report(missing_msg, fmt_kwargs)

    # get node at index
    stu_part = stu_out[index]
    sol_part = sol_out[index]

    append_message = FeedbackComponent(expand_msg, fmt_kwargs)

    return part_to_child(stu_part, sol_part, append_message, state, node_name=name)


# context functions -----------------------------------------------------------
# TODO: check if still useful
def with_context(state, *args, child=None):

    # set up context in processes
    solution_res = setUpNewEnvInProcess(
        process=state.solution_process, context=state.solution_parts["with_items"]
    )
    if isinstance(solution_res, Exception):
        with debugger(state):
            state.report(
                "ошибка в решении при вызове test_with(): %s" % str(solution_res)
            )

    student_res = setUpNewEnvInProcess(
        process=state.student_process, context=state.student_parts["with_items"]
    )
    if isinstance(student_res, AttributeError):
        child.report(
            "В вашей конструкция 'with', вы не используете корректный контекстный менеджер."
        )

    if isinstance(student_res, (AssertionError, ValueError, TypeError)):
        child.report(
            "В вашей конструкция 'with', количество значений в вашем контекстном менеджере "
            "не соответствует количеству переменных, которым вы пытаетесь их назначить."
        )

    # run subtests
    try:
        multi(state, *args)
    finally:
        # exit context
        close_solution_context = breakDownNewEnvInProcess(
            process=state.solution_process
        )
        if isinstance(close_solution_context, Exception):
            with debugger(state):
                state.report(
                    "error in the solution, closing the `with` fails with: %s"
                    % close_solution_context
                )

        close_student_context = breakDownNewEnvInProcess(process=state.student_process)
        if isinstance(close_student_context, Exception):
            state.report(
                "Конструкция 'with' не может быть закрыта корретно, вы "
                "используете контекстный менеджер неправильно."
            )
    return state


def check_args(state, name, missing_msg=None):
    """Check whether a function argument is specified.

    This function can follow ``check_function()`` in an SCT chain and verifies whether an argument is specified.
    If you want to go on and check whether the argument was correctly specified, you can can continue chaining with
    ``has_equal_value()`` (value-based check) or ``has_equal_ast()`` (AST-based check)

    This function can also follow ``check_function_def()`` or ``check_lambda_function()`` to see if arguments have been
    specified.

    Args:
        name (str): the name of the argument for which you want to check if it is specified. This can also be
            a number, in which case it refers to the positional arguments. Named arguments take precedence.
        missing_msg (str): If specified, this overrides the automatically generated feedback message in case
            the student did specify the argument.
        state (State): State object that is passed from the SCT Chain (don't specify this).

    :Examples:

        Student and solution code::

            import numpy as np
            arr = np.array([1, 2, 3, 4, 5])
            np.mean(arr)

        SCT::

            # Verify whether arr was correctly set in np.mean
            # has_equal_value() checks the value of arr, used to set argument a
            Ex().check_function('numpy.mean').check_args('a').has_equal_value()

            # Verify whether arr was correctly set in np.mean
            # has_equal_ast() checks the expression used to set argument a
            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()

        Student and solution code::

            def my_power(x):
                print("calculating sqrt...")
                return(x * x)

        SCT::

            Ex().check_function_def('my_power').multi(
                check_args('x') # will fail if student used y as arg
                check_args(0)   # will still pass if student used y as arg
            )

    """
    if missing_msg is None:
        missing_msg = "Вы указали {{part}}?"

    if name in ["*args", "**kwargs"]:  # for check_function_def
        return check_part(state, name, name, missing_msg=missing_msg)
    else:
        if isinstance(name, list):  # dealing with args or kwargs
            if name[0] == "args":
                arg_str = "{} аргумент передается как аргумент переменной длины".format(
                    get_ord(name[1] + 1)
                )
            else:
                arg_str = "argument `{}`".format(name[1])
        else:
            arg_str = (
                "{} argument".format(get_ord(name + 1))
                if isinstance(name, int)
                else "argument `{}`".format(name)
            )
        return check_part_index(state, "args", name, arg_str, missing_msg=missing_msg)


# CALL CHECK ==================================================================


def build_call(callstr, node):
    if isinstance(node, ast.FunctionDef):  # function name
        func_expr = ast.Name(id=node.name, ctx=ast.Load())
        argstr = "`%s`" % callstr.replace("f", node.name)
    elif isinstance(node, ast.Lambda):  # lambda body expr
        func_expr = node
        argstr = "it with the arguments `%s`" % callstr.replace("f", "")
    else:
        raise TypeError("Не удается обработать переданный AST.")

    parsed = ast.parse(callstr).body[0].value
    parsed.func = func_expr
    ast.fix_missing_locations(parsed)
    return parsed, argstr


def check_call(state, callstr, argstr=None, expand_msg=None):
    """When checking a function definition of lambda function,
    prepare has_equal_x for checking the call of a user-defined function.

    Args:
        callstr (str): call string that specifies how the function should be called, e.g. `f(1, a = 2)`.
           ``check_call()`` will replace ``f`` with the function/lambda you're targeting.
        argstr (str): If specified, this overrides the way the function call is refered to in the expand message.
        expand_msg (str): If specified, this overrides any messages that are prepended by previous SCT chains.
        state (State): state object that is chained from.

    :Example:

        Student and solution code::

            def my_power(x):
                print("calculating sqrt...")
                return(x * x)

        SCT::

            Ex().check_function_def('my_power').multi(
                check_call("f(3)").has_equal_value()
                check_call("f(3)").has_equal_output()
            )
    """

    state.assert_is(
        ["function_defs", "lambda_functions"],
        "check_call",
        ["check_function_def", "check_lambda_function"],
    )

    if expand_msg is None:
        expand_msg = "To verify it, we reran {{argstr}}. "

    stu_part, _argstr = build_call(callstr, state.student_parts["node"])
    sol_part, _ = build_call(callstr, state.solution_parts["node"])

    append_message = FeedbackComponent(expand_msg, {"argstr": argstr or _argstr})
    child = part_to_child(stu_part, sol_part, append_message, state)

    return child



from tcs_pythonwhat.State import State
from tcs_pythonwhat.local import run_exercise
from tcs_pythonwhat.sct_syntax import Ex, get_chains
from tcs_pythonwhat.utils import check_str, check_process
from protowhat.Reporter import Reporter
from protowhat.failure import Failure, InstructorError
from tcs_pythonwhat.utils import include_v1


def test_exercise(
    sct,
    student_code,
    solution_code,
    pre_exercise_code,
    student_process,
    solution_process,
    raw_student_output,
    ex_type,
    error,
    force_diagnose=False,
):
    """
    Point of interaction with the Python backend.
    Args:
            sct (str): The solution corectness test as a string of code.
            student_code (str): The code which is entered by the student.
            solution_code (str): The code which is in the solution.
            pre_exercise_code (str): The code which is executed pre exercise.
            student_process (Process): Process in which the student code was executed.
            solution_process (Process): Process in which the solution code was executed.
            raw_student_output (str): The output which is given by executing the student's program.
            ex_type (str): The type of the exercise.
            error (tuple): A tuple with some information on possible errors.
    Returns:
            dict: Returns dict with correct - whether the SCT passed, message - the feedback message and
              tags - the tags belonging to the SCT execution.
    """

    reporter = Reporter(errors=[error] if error else [])

    try:
        state = State(
            student_code=check_str(student_code),
            solution_code=check_str(solution_code),
            pre_exercise_code=check_str(pre_exercise_code),
            student_process=check_process(student_process),
            solution_process=check_process(solution_process),
            raw_student_output=check_str(raw_student_output),
            force_diagnose=force_diagnose,
            reporter=reporter,
        )

        State.root_state = state
        tree, sct_cntxt = prep_context()

        # Actually execute SCTs
        exec(sct, sct_cntxt)

        # Run remaining nodes on tree (v1 only)
        if tree:
            for test in tree.crnt_node:
                test(state)

    except Failure as e:
        if isinstance(e, InstructorError):
            # TODO: decide based on context
            raise e
        return reporter.build_failed_payload(e.feedback)

    return reporter.build_final_payload()


# TODO: consistent success_msg
def success_msg(message):
    """
    Set the succes message of the sct. This message will be the feedback if all tests pass.
    Args:
            message (str): A string containing the feedback message.
    """
    State.root_state.reporter.success_msg = message


# deprecated
def allow_errors():
    State.root_state.reporter.errors_allowed = True


def prep_context():
    cntxt = {"success_msg": success_msg}
    from pythonwhat.sct_syntax import v2_check_functions
    from pythonwhat.probe import build_probe_context

    imports = [
        "from inspect import Parameter as param",
        "from pythonwhat.signatures import sig_from_params, sig_from_obj",
        "from pythonwhat.State import set_converter",
        "from pythonwhat.sct_syntax import F, Ex"
    ]
    [exec(line, None, cntxt) for line in imports]

    # only if PYTHONWHAT_V2_ONLY is not set, support v1
    if include_v1():
        tree, probe_cntxt = build_probe_context()
        cntxt.update(probe_cntxt)
    else:
        tree = None

    cntxt.update(v2_check_functions)
    # TODO: ChainStart instances cause errors when dill tries to pass manual converter functions
    # cntxt.update(get_chains())
    return tree, cntxt


def setup_state(stu_code="", sol_code="", pec="", **kwargs):
    sol_process, stu_process, raw_stu_output, error = run_exercise(
        pec, sol_code, stu_code, **kwargs
    )

    state = State(
        student_code=stu_code,
        solution_code=sol_code,
        pre_exercise_code=pec,
        student_process=stu_process,
        solution_process=sol_process,
        raw_student_output=raw_stu_output,
        reporter=Reporter(errors=[error] if error else []),
    )

    State.root_state = state
    return Ex(state)



from protowhat.failure import _debug
from protowhat.checks.check_simple import allow_errors
from protowhat.checks.check_bash_history import has_command
from protowhat.checks.check_files import check_file, has_dir
from tcs_pythonwhat.checks.check_funcs import check_part, check_part_index, check_node
from tcs_pythonwhat.checks.has_funcs import has_equal_part
from tcs_pythonwhat.checks.check_function import check_function
from tcs_pythonwhat.checks.check_has_context import has_context
from tcs_pythonwhat.checks import check_object, check_logic, check_funcs, has_funcs
from tcs_pythonwhat.local import run

from inspect import signature, Parameter
from functools import partial, wraps
from jinja2 import Template

__PART_WRAPPERS__ = {
    "iter": "iterable part",
    "body": "body",
    "key": "key part",
    "value": "value part",
    "orelse": "else part",
    "finalbody": "finally part",
    "test": "condition",
}

__PART_INDEX_WRAPPERS__ = {
    "ifs": "{{ordinal}} if",
    "bases": "{{ordinal}} base class",
    "handlers": "`{{index}}` `except` block",
    "context": "{{ordinal}} context",
}

__NODE_WRAPPERS__ = {
    "list_comp": {
        "typestr": "{{ordinal}} list comprehension",
        "docstr": """Check whether a list comprehension was coded and zoom in on it.

        Can be chained with ``check_iter()``, ``check_body()``, and ``check_ifs()``.

        Args:
            index: Index of the list comprehension (0-based)
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you expect students to create a list ``my_list`` as follows: ::

                my_list = [ i*2 for i in range(0,10) if i>2 ]

            The following SCT would robustly verify this: ::

                Ex().check_correct(
                    check_object('my_list').has_equal_value(),
                    check_list_comp().multi(
                        check_iter().has_equal_value(),
                        check_body().set_context(4).has_equal_value(),
                        check_ifs(0).multi(
                            set_context(0).has_equal_value(),
                            set_context(3).has_equal_value(),
                            set_context(5).has_equal_value()
                        )
                    )
                )

            - With ``check_correct()``, we're making sure that the list comprehension
              checking is not executed if ``my_list`` was calculated properly.
            - If ``my_list`` is not correct, the 'diagnose' chain will run: ``check_list_comp()`` looks
              for the first list comprehension in the student's submission.
            - Next, ``check_iter()`` zooms in on the iterator, ``range(0, 10)`` in the case of the solution.
              ``has_equal_value()`` verifies whether the expression that the student used evaluates to the
              same value as the expression that the solution used.
            - ``check_body()`` zooms in on the body, ``i*2`` in the case of the solution.
              ``set_context()`` sets the iterator to 4, allowing for the fact that the student used another name instead of ``i`` for this iterator.
              ``has_equal_value()`` reruns the body in the student and solution code with the iterator set to 4, and checks if the results are the same.
            - ``check_ifs(0)`` zooms in on the first ``if`` of the list comprehension, ``i>2`` in case of the solution.
              With a series of ``set_context()`` and ``has_equal_value()``, it is verifies whether this condition evaluates to the same value in student
              and solution code for different values of the iterator (`i` in the case of the solution, whatever in the case of the student).

        """,
    },
    "generator_exp": {
        "typestr": "{{ordinal}} generator expression",
        "docstr": """Check whether a generator expression was coded and zoom in on it.

        Can be chained with ``check_iter()``, ``check_body()``, and ``check_ifs()``.

        Args:
            index: Index of the generator expression (0-based)
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you expect students to create a generator ``my_gen`` as follows: ::

                my_gen = ( i*2 for i in range(0,10) )

            The following SCT would robustly verify this: ::

                Ex().check_correct(
                    check_object('my_gen').has_equal_value(),
                    check_generator_exp().multi(
                        check_iter().has_equal_value(),
                        check_body().set_context(4).has_equal_value()
                    )
                )

            Have a look at ``check_list_comp`` to understand what's going on; it is very similar.

        """,
    },
    "dict_comp": {
        "typestr": "{{ordinal}} dictionary comprehension",
        "docstr": """Check whether a dictionary comprehension was coded and zoom in on it.

        Can be chained with ``check_key()``, ``check_value()``, and ``check_ifs()``.

        Args:
            index: Index of the dictionary comprehension (0-based)
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you expect students to create a dictionary ``my_dict`` as follows: ::

                my_dict = { m:len(m) for m in ['a', 'ab', 'abc'] }

            The following SCT would robustly verify this: ::

                Ex().check_correct(
                    check_object('my_dict').has_equal_value(),
                    check_dict_comp().multi(
                        check_iter().has_equal_value(),
                        check_key().set_context('ab').has_equal_value(),
                        check_value().set_context('ab').has_equal_value()
                    )
                )

            - With ``check_correct()``, we're making sure that the dictionary comprehension
              checking is not executed if ``my_dict`` was created properly.
            - If ``my_dict`` is not correct, the 'diagnose' chain will run: ``check_dict_comp()`` looks
              for the first dictionary comprehension in the student's submission.
            - Next, ``check_iter()`` zooms in on the iterator, ``['a', 'ab', 'abc']`` in the case of the solution.
              ``has_equal_value()`` verifies whether the expression that the student used evaluates to the
              same value as the expression that the solution used.
            - ``check_key()`` zooms in on the key of the comprehension, ``m`` in the case of the solution.
              ``set_context()`` temporaritly sets the iterator to ``'ab'``, allowing for the fact that the student used another name instead of ``m`` for this iterator.
              ``has_equal_value()`` reruns the key expression in the student and solution code with the iterator set to ``'ab'``, and checks if the results are the same.
            - ``check_value()`` zooms in on the value of the comprehension, ``len(m)`` in the case of the solution.
              ``has_equal_value()`` reruns the value expression in the student and solution code with the iterator set to ``'ab'``, and checks if the results are the same.

        """,
    },
    "for_loop": {
        "typestr": "{{ordinal}} for loop",
        "docstr": """Check whether a for loop was coded and zoom in on it.

        Can be chained with ``check_iter()`` and ``check_body()``.

        Args:
            index: Index of the for loop (0-based).
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want a student to iterate over a predefined dictionary ``my_dict`` and do the appropriate printouts: ::

                for key, value in my_dict.items():
                    print(key + " - " + str(value))

            The following SCT would verify this: ::

                Ex().check_for_loop().multi(
                    check_iter().has_equal_value(),
                    check_body().multi(
                        set_context('a', 1).has_equal_output(),
                        set_context('b', 2).has_equal_output()
                    )
                )

            - ``check_for_loop()`` zooms in on the ``for`` loop, and makes its parts available for further checking.
            - ``check_iter()`` zooms in on the iterator part of the for loop, ``my_dict.items()`` in the solution.
              ``has_equal_value()`` re-executes the expressions specified by student and solution and compares their results.
            - ``check_body()`` zooms in on the body part of the for loop, ``print(key + " - " + str(value))``.
              For different values of ``key`` and ``value``, the student's body and solution's body are executed again and the printouts are captured and compared to see if they are equal.

            Notice how you do not need to specify the variables by name in ``set_context()``. pythonwhat can figure out the variable names used in both student and solution code, and
            can do the verification independent of that. That way, we can make the SCT robust against submissions that code the correct logic, but use different names for the context values.
            In other words, the following student submissions that would also pass the SCT: ::

                # passing submission 1
                my_dict = {'a': 1, 'b': 2}
                for k, v in my_dict.items():
                    print(k + " - " + str(v))

                # passing submission 2
                my_dict = {'a': 1, 'b': 2}
                for first, second in my_dict.items():
                    mess = first + " - " + str(second)
                    print(mess)

        :Example:

            As another example, suppose you want the student to build a list of doubles as follows: ::

                even = []
                for i in range(10):
                    even.append(2*i)

            The following SCT would robustly verify this: ::

                Ex().check_correct(
                    check_object('even').has_equal_value(),
                    check_for_loop().multi(
                        check_iter().has_equal_value(),
                        check_body().set_context(2).set_env(even = []).\\
                            has_equal_value(name = 'even')
                    )
                )

            - ``check_correct()`` makes sure that we do not dive into the ``for`` loop if the array ``even`` is correctly populated in the end.
            - If ``even`` was not correctly populated, ``check_for_loop()`` will zoom in on the for loop.
            - The ``check_iter()`` chain verifies whether `range(10)` (or something equivalent) was used to iterate over.
            - ``check_body()`` zooms in on the body, and reruns the body (``even.append(2*i)`` in the solution) for ``i`` equal to 2, and even temporarily set to an empty array.
              Notice how we use ``set_context()`` to robustly set the context value (the student can use a different variable name), while we have to explicitly set ``even`` with ``set_env()``.
              Also notice how we use ``has_equal_value(name = 'even')`` instead of the usual ``check_object()``; ``check_object()`` can only be called from the root state ``Ex()``.

        :Example:

            As a follow-up example, suppose you want the student to build a list of doubles of the even numbers only: ::

                even = []
                for i in range(10):
                    if i % 2 == 0:
                        even.append(2*i)

            The following SCT would robustly verify this: ::

                Ex().check_correct(
                    check_object('even').has_equal_value(),
                    check_for_loop().multi(
                        check_iter().has_equal_value(),
                        check_body().check_if_else().multi(
                            check_test().multi(
                                set_context(1).has_equal_value(),
                                set_context(2).has_equal_value()
                            ),
                            check_body().set_context(2).\\
                                set_env(even = []).has_equal_value(name = 'even')
                        )
                    )
                )

        """,
    },
    "function_def": {
        "typestr": "definition of `{{index}}()`",
        "docstr": """Check whether a function was defined and zoom in on it.

        Can be chained with ``check_call()``, ``check_args()`` and ``check_body()``.

        Args:
            index: the name of the function definition.
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want a student to create a function ``shout_echo()``: ::

                def shout_echo(word1, echo=1):
                    echo_word = word1 * echo
                    shout_words = echo_word + '!!!'
                    return shout_words

            The following SCT robustly checks this: ::

                Ex().check_function_def('shout_echo').check_correct(
                    multi(
                        check_call("f('hey', 3)").has_equal_value(),
                        check_call("f('hi', 2)").has_equal_value(),
                        check_call("f('hi')").has_equal_value()
                    ),
                    check_body().set_context('test', 1).multi(
                        has_equal_value(name = 'echo_word'),
                        has_equal_value(name = 'shout_words')
                    )
                )

            Here:

            - ``check_function_def()`` zooms in on the function definition of ``shout_echo`` in both student and solution code (and process).
            - ``check_correct()`` is used to

                + First check whether the function gives the correct result when called in different ways (through ``check_call()``).
                + Only if these 'function unit tests' don't pass, ``check_correct()`` will run the `check_body()` chain that dives deeper into the
                  function definition body. This chain sets the context variables - ``word1`` and ``echo``, the arguments of the function - to
                  the values ``'test'`` and ``1`` respectively, again while being agnostic to the actual name of these context variables.

            Notice how ``check_correct()`` is used to great effect here: why check the function definition internals if the I/O of the function works fine?
            Because of this construct, all the following submissions will pass the SCT: ::

                # passing submission 1
                def shout_echo(w, e=1):
                    ew = w * e
                    return ew + '!!!'

                # passing submission 2
                def shout_echo(a, b=1):
                    return a * b + '!!!'

        :Example:

            ``check_args()`` is most commonly used in combination with ``check_function()``
            to verify the arguments of function **calls**, but it can also be used
            to verify the arguments specified in the signature of a function definition.

            We can extend the SCT for the previous example to explicitly verify the signature: ::


                msg1 = "Make sure to specify 2 arguments!"
                msg2 = "don't specify default arg!"
                msg3 = "specify a default arg!"
                Ex().check_function_def('shout_echo').check_correct(
                    multi(
                        check_call("f('hey', 3)").has_equal_value(),
                        check_call("f('hi', 2)").has_equal_value(),
                        check_call("f('hi')").has_equal_value()
                    ),
                    multi(
                        has_equal_part_len("args", unequal_msg=1),
                        check_args(0).has_equal_part('is_default', msg=msg2),
                        check_args('word1').has_equal_part('is_default', msg=msg2),
                        check_args(1).\\
                            has_equal_part('is_default', msg=msg3).has_equal_value(),
                        check_args('echo').\\
                            has_equal_part('is_default', msg=msg3).has_equal_value(),
                        check_body().set_context('test', 1).multi(
                            has_equal_value(name = 'echo_word'),
                            has_equal_value(name = 'shout_words')
                        )
                    )
                )

            - ``has_equal_part_len("args")`` verifies whether student and solution function
              definition have the same number of arguments.
            - ``check_args(0)`` refers to the first argument in the signature by position,
              and the chain checks whether the student did not specify a default as in the solution.
            - An alternative for the ``check_args(0)`` chain is to use ``check_args('word1')``
              to refer to the first argument. This is more restrictive, as the requires the
              student to use the exact same name.
            - ``check_args(1)`` refers to the second argument in the signature by position,
              and the chain checks whether the student specified a default, as in the solution, and
              whether the value of this default corresponds to the one in the solution.
            - The ``check_args('echo')`` chain is a more restrictive alternative for the ``check_args(1)``
              chain.

            Notice that support for verifying arguments is not great yet:

            - A lot of work is needed to verify the number of arguments and whether or not defaults are set.
            - You have to specify custom messages because pythonwhat doesn't automatically generate messages.

            We are working on it!

        """,
    },
    "class_def": {
        "typestr": "class definition of `{{index}}`",
        "docstr": """Check whether a class was defined and zoom in on its definition

        Can be chained with ``check_bases()`` and ``check_body()``.

        Args:
            index: the name of the function definition.
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want to check whether a class was defined correctly: ::

                class MyInt(int):
                    def __init__(self, i):
                        super().__init__(i + 1)

            The following SCT would verify this: ::

                Ex().check_class_def('MyInt').multi(
                    check_bases(0).has_equal_ast(),
                    check_body().check_function_def('__init__').multi(
                        check_args('self'),
                        check_args('i'),
                        check_body().set_context(i = 2).multi(
                            check_function('super', signature=False),
                            check_function('super.__init__').check_args(0).has_equal_value()
                        )
                    )
                )

            - ``check_class_def()`` looks for the class definition itself.
            - With ``check_bases()``, you can zoom in on the different basse classes that the class definition inherits from.
            - With ``check_body()``, you zoom in on the class body, after which you can use other functions such
              as ``check_function_def()`` to look for class methods.
            - Of course, just like for other examples, you can use ``check_correct()`` where necessary,
              e.g. to verify whether class methods give the right behavior with ``check_call()``
              before diving into the body of the method itself.

        """,
    },
    "if_exp": {
        "typestr": "{{ordinal}} if expression",
        "docstr": """Check whether an if expression was coded zoom in on it.

        This function works the exact same way as ``check_if_else()``.
        """,
    },
    "if_else": {
        "typestr": "{{ordinal}} if statement",
        "docstr": """Check whether an if statement was coded zoom in on it.

        Args:
            index: the index of the if statement to look for (0 based)
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want students to print out a message if ``x`` is larger than 0: ::

                x = 4
                if x > 0:
                    print("x is strictly positive")

            The following SCT would verify that: ::

                Ex().check_if_else().multi(
                    check_test().multi(
                        set_env(x = -1).has_equal_value(),
                        set_env(x =  1).has_equal_value(),
                        set_env(x =  0).has_equal_value()
                    ),
                    check_body().check_function('print', 0).\\
                        check_args('value').has_equal_value()
                )

            - ``check_if_else()`` zooms in on the first if statement in the student and solution submission.
            - ``check_test()`` zooms in on the 'test' portion of the if statement, ``x > 0`` in case of the solution.
              ``has_equal_value()`` reruns this expression and the corresponding expression in the student code for
              different values of ``x`` (set with ``set_env()``) and compare there results.
              This way, you can robustly verify whether the if test was coded up correctly. If the student
              codes up the condition as ``0 < x``, this would also be accepted.
            - ``check_body()`` zooms in on the 'body' portion of the if statement, ``print("...")`` in case of the solution.
              With a classical ``check_function()`` chain, it is verified whether the if statement contains a
              function ``print()`` and whether its argument is set correctly.

        :Example:

            In Python, when an if-else statement has an ``elif`` clause, it is held in the `orelse` part.
            In this sense, an if-elif-else statement is represented by python as nested if-elses.
            More specifically, this if-else statement: ::

                if x > 0:
                    print(x)
                elif y > 0:
                    print(y)
                else:
                    print('none')

            Is syntactically equivalent to: ::

                if x > 0:
                    print(x)
                else:
                    if y > 0:
                        print(y)
                    else:
                        print('none')

            The second representation has to be followed when writing the corresponding SCT: ::

                Ex().check_if_else().multi(
                    check_test(),          # zoom in on x > 0
                    check_body(),          # zoom in on print(x)
                    check_orelse().check_if_else().multi(
                        check_test(),      # zoom in on y > 0
                        check_body(),      # zoom in on print(y)
                        check_orelse()     # zoom in on print('none')
                    )
                )

        """,
    },
    "lambda_function": {
        "typestr": "{{ordinal}} lambda function",
        "docstr": """Check whether a lambda function was coded zoom in on it.

        Can be chained with ``check_call()``, ``check_args()`` and ``check_body()``.

        Args:
            index: the index of the lambda function (0-based).
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want a student to create a lambda function
            that returns the length of an array times two: ::

                lambda x: len(x)*2

            The following SCT robustly checks this: ::

                Ex().check_lambda_function().check_correct(
                    multi(
                        check_call("f([1])").has_equal_value(),
                        check_call("f([1, 2])").has_equal_value()
                    ),
                    check_body().set_context([1, 2, 3]).has_equal_value()
                )

            Here:

            - ``check_lambda_function()`` zooms in on the first lambda function in both student and solution code.
            - ``check_correct()`` is used to

                + First check whether the lambda function gives the correct result when called in different ways (through ``check_call()``).
                + Only if these 'function unit tests' don't pass, ``check_correct()`` will run the `check_body()` chain that dives deeper into the
                  lambda function's body. This chain sets the context variable `x`, the argument of the function, to
                  the values ``[1, 2, 3]``, while being agnostic to the actual name the student used for this context variable.

            Notice how ``check_correct()`` is used to great effect here: why check the function definition internals if the I/O of the function works fine?
            Because of this construct, all the following submissions will pass the SCT: ::

                # passing submission 1
                lambda x: len(x) + len(x)

                # passing submission 2
                lambda y, times=2: len(y) * times
        """,
    },
    "try_except": {
        "typestr": "{{ordinal}} try statement",
        "docstr": """Check whether a try except statement was coded zoom in on it.

        Can be chained with ``check_body()``, ``check_handlers()``, ``check_orelse()`` and ``check_finalbody()``.

        Args:
            index: the index of the try except statement (0-based).
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want to verify whether the student did a `try-except` statement properly: ::

                do_dangerous_thing = lambda n: n

                try:
                    x = do_dangerous_thing(n = 4)
                except ValueError as e:
                    x = 'something wrong with inputs'
                except:
                    x = 'something went wrong'
                finally:
                    print('ciao!')

            The following SCT can be used to verify this: ::

                Ex().check_try_except().multi(
                    check_body().\\
                        check_function('do_dangerous_thing').\\
                        check_args('n').has_equal_value(),
                    check_handlers('ValueError').\\
                        has_equal_value(name = 'x'),
                    check_handlers('all').\\
                        has_equal_value(name = 'x'),
                    check_finalbody().\\
                        check_function('print').check_args(0).has_equal_value()
                )

        """,
    },
    "while": {
        "typestr": "{{ordinal}} `while` loop",
        "docstr": """Check whether a while loop was coded and zoom in on it.

        Can be chained with ``check_test()``, ``check_body()`` and ``check_orelse()``.

        Args:
            index: the index of the while loop to verify (0-based).
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        :Example:

            Suppose you want a student to code a while loop that counts down a counter from 50 until
            a multilpe of 11 is found. If it is found, the value should be printed out. ::

                i = 50
                while i % 11 != 0:
                    i -= 1

            The following SCT robustly verifies this: ::

                Ex().check_correct(
                    check_object('i').has_equal_value(),
                    check_while().multi(
                            check_test().multi(
                                set_env(i = 45).has_equal_value(),
                                set_env(i = 44).has_equal_value()
                            ),
                            check_body().set_env(i = 3).has_equal_value(name = 'i')
                    )
                )

            - ``check_correct()`` first checks whether the end result of ``i`` is correct. If it is, the entire chain that checks the ``while`` loop is skipped.
            - If ``i`` is not correctly calculated, ``check_while_loop()`` zooms in on the while loop.
            - ``check_test()`` zooms in on the condition of the ``while`` loop, ``i % 11 != 0`` in the solution, and verifies whether
              the expression gives the same results for different values of ``i``, set through ``set_env()``, when comparing student and solution.
            - ``check_body()`` zooms in on the body of the ``while`` loop, and ``has_equal_value()`` checks whether rerunning this body
              updates ``i`` as expected when ``i`` is temporarily set to 3 with ``set_env()``.

        """,
    },
    "with": {
        "typestr": "{{ordinal}} `with` statement",
        "docstr": """Check whether a with statement was coded zoom in on it.

        Args:
            index: the index of the``with`` statement to verify (0-based)
            {{typestr}}
            {{missing_msg}}
            {{expand_msg}}

        """,
    },
}

scts = dict()

# make has_equal_part wrappers


def partial_with_offset(offset=1):
    def bound_partial_with_offset(func, *partial_args, **partial_kwargs):
        kwargs_partial = partial(func, **partial_kwargs)

        @wraps(func)
        def full_partial(*args, **kwargs):
            full_args = args[:offset] + partial_args + args[offset:]
            return kwargs_partial(*full_args, **kwargs)

        # set correct signature of returned partial
        # todo: pass arguments as keywords to partial, instead of this decorator?
        #  (where args are always the same)
        func_sig = signature(full_partial)
        parameter_names = tuple(func_sig.parameters)

        partialed_positional_indices = []
        for kwarg in partial_kwargs:
            param = func_sig.parameters[kwarg]
            if param.default is param.empty:
                partialed_positional_indices.append(parameter_names.index(kwarg))

        partial_params = list(func_sig.parameters.values())
        for index in sorted(partialed_positional_indices, reverse=True):
            # appending isn't needed for functionality, but more similar to partial
            # and it shows that these arguments can still be updated as kwargs
            partial_params.append(
                partial_params[index].replace(
                    kind=Parameter.KEYWORD_ONLY,
                    default=partial_kwargs[partial_params[index].name],
                )
            )
            del partial_params[index]
        del partial_params[offset : offset + len(partial_args)]

        full_partial.__signature__ = func_sig.replace(parameters=partial_params)

        return full_partial

    return bound_partial_with_offset


state_partial = partial_with_offset()


def rename_function(func, name):
    # see functools.wraps
    func.__name__ = func.__qualname__ = name


def add_partial_sct(func, name):
    rename_function(func, name)
    scts[name] = func


add_partial_sct(
    state_partial(
        has_equal_part,
        "name",
        msg="Убедитесь, что вы корретно используете {{name}}, ожидалось {{sol_part[name]}}, но было получено {{stu_part[name]}}.",
    ),
    "has_equal_name",
)
add_partial_sct(
    state_partial(
        has_equal_part,
        "is_default",
        msg="Убедитесь, что {{ 'has' if sol_part.is_default else 'does not have'}} является аргументом по-умолчанию.",
    ),
    "is_default",
)

# include rest of wrappers
for k, v in __PART_WRAPPERS__.items():
    check_fun = state_partial(check_part, k, v)
    add_partial_sct(check_fun, "check_" + k)

for k, v in __PART_INDEX_WRAPPERS__.items():
    check_fun = state_partial(check_part_index, k, part_msg=v)
    add_partial_sct(check_fun, "check_" + k)

for k, v in __NODE_WRAPPERS__.items():
    check_fun = state_partial(check_node, k + "s", typestr=v["typestr"])
    check_fun.__doc__ = Template(v["docstr"]).render(
        typestr="typestr: If specified, this overrides the standard way of referring to the construct you're zooming in on.",
        missing_msg="missing_msg: If specified, this overrides the automatically generated feedback message in case the construct could not be found.",
        expand_msg="expand_msg: If specified, this overrides the automatically generated feedback message that is prepended to feedback messages that are thrown further in the SCT chain.",
    )
    add_partial_sct(check_fun, "check_" + k)

for k in [
    "set_context",
    "set_env",
    "disable_highlighting",
    "check_not",
    "check_or",
    "check_correct",
    "fail",
    "override",
    "multi",
]:
    scts[k] = getattr(check_logic, k)

for k in ["with_context", "check_args", "check_call"]:
    scts[k] = getattr(check_funcs, k)

for k in [
    "has_equal_value",
    "has_equal_output",
    "has_equal_error",
    "has_equal_ast",
    "has_equal_part_len",
    "has_equal_part",
    "has_import",
    "has_output",
    "has_printout",
    "has_code",
    "has_no_error",
    "has_chosen",
]:
    sct = getattr(has_funcs, k)
    if not hasattr(sct, "__name__"):
        rename_function(sct, k)
    scts[k] = sct

# include check_object and friends ------
for k in ["check_object", "is_instance", "check_df", "check_keys"]:
    scts[k] = getattr(check_object, k)

scts["has_context"] = has_context
scts["check_function"] = check_function

scts["run"] = run

scts["has_command"] = has_command

scts["check_file"] = check_file
scts["has_dir"] = has_dir

scts["allow_errors"] = allow_errors

scts["_debug"] = _debug

locals().update(scts)



from .test_compound_statement import (
    test_with,
    test_list_comp,
    test_if_else,
    test_for_loop,
    test_while_loop,
    test_expression_output,
    test_expression_result,
    test_object_after_expression,
    test_function_definition,
)

from .test_object import test_object, test_data_frame
from .test_function import test_function, test_function_v2
from .test_object_accessed import test_object_accessed

from pythonwhat.checks.check_logic import (  # noqa
    check_or as test_or,
    check_correct as test_correct,
)

from pythonwhat.checks.has_funcs import (  # noqa
    has_code as test_student_typed,
    has_import as test_import,
    has_output as test_output_contains,
    has_chosen as test_mc,
)



import io
import os
import random
from pathlib import Path
from contextlib import redirect_stdout

from multiprocessing import Process, Queue
from protowhat.Reporter import Reporter

try:
    from pythonbackend.shell_utils import create
    from pythonbackend.tasks import TaskCaptureFullOutput

    BACKEND_AVAILABLE = True
except:
    BACKEND_AVAILABLE = False


class StubShell:
    def __init__(self, init_code=None):
        self.user_ns = {}
        if init_code:
            self.run_code(init_code)

    def run_code(self, code):
        exec(code, self.user_ns)


class StubProcess:
    def __init__(self, init_code=None, pid=None):
        self.shell = StubShell(init_code)
        self._identity = (pid,) if pid else (random.randint(0, 1e12),)

    def executeTask(self, task):
        return task(self.shell)


class TaskCaptureOutput:
    def __init__(self, code):
        self.code = code

    def __call__(self, shell):
        return run_code(shell.run_code, self.code)


class TaskKillProcess:
    def __call__(self, shell):
        return None


class CaptureErrors:
    def __init__(self, output):
        self.output = output

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exception, traceback):
        if exc_type is not None:
            self.output.append({"type": "backend-error", "payload": str(exception)})
            return True


class WorkerProcess(Process):
    instances = []

    def __init__(self, pid=None):
        Process.__init__(self)
        self.task_queue = Queue()
        self.result_queue = Queue()
        self.daemon = (
            True
        )  # when parent process is killed, sub/childprocess get also killed
        self.instances.append(self)
        # used to detect single process exercise
        self._identity = (pid,) if pid else (random.randint(0, 1e12),)

    def get_shell(self):
        return create({})

    def run(self):
        shell = self.get_shell()
        while True:
            output = []
            with CaptureErrors(output):
                next_task = self.task_queue.get()
                answer = next_task(shell)
            if len(output) > 0:  # means backend error happened
                answer = output
            output = []
            with CaptureErrors(output):
                self.result_queue.put_nowait(answer)
            if len(output) > 0:  # means backend error happened
                self.result_queue.put_nowait(output)
            if isinstance(next_task, TaskKillProcess):
                break  # break while loop -> we do not wait upon new task
        return

    def executeTask(self, task):
        self.task_queue.put_nowait(task)
        return self.result_queue.get()  # wait and fetches next item in queue

    def kill(self):
        try:
            if self.is_alive():
                self.executeTask(TaskKillProcess())
                self.join(timeout=3.0)
                if self.is_alive():
                    self.terminate()
                    self.join(timeout=3.0)
            if self in self.instances:
                self.instances.remove(self)
        finally:
            pass
            # python 3.7:
            # self.close()

    @classmethod
    def kill_all(cls):
        for instance in list(cls.instances):
            instance.kill()


class SimpleProcess(WorkerProcess):
    def get_shell(self):
        return StubShell()


class ChDir(object):
    """
    Step into a directory temporarily.
    """

    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = str(path)

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)


def run_code(executor, code):
    with io.StringIO() as output:
        try:
            with redirect_stdout(output):
                executor(code)
            raw_output = output.getvalue()
            error = None
        except BaseException as e:
            raw_output = ""
            error = str(e)
    return raw_output, error


def run_single_process(pec, code, pid=None, mode="simple"):
    if mode == "stub":
        # no isolation
        process = StubProcess(init_code=pec, pid=pid)
        raw_output, error = run_code(process.shell.run_code, code)

    elif mode == "simple":
        # no advanced functionality
        process = SimpleProcess(pid)
        process.start()
        _ = process.executeTask(TaskCaptureOutput(pec))
        raw_output, error = process.executeTask(TaskCaptureOutput(code))

    elif mode == "full" and BACKEND_AVAILABLE:
        # slow
        process = WorkerProcess(pid)
        process.start()
        _ = process.executeTask(
            TaskCaptureFullOutput((pec,), "<PEC>", None, silent=True)
        )
        output, raw_output = process.executeTask(
            TaskCaptureFullOutput((code,), "script.py", None, silent=True)
        )
        raw_output = raw_output["output_stream"]
        error = raw_output["error"]

    else:
        raise ValueError("Invalid mode")

    return process, raw_output, error


def run_exercise(pec, sol_code, stu_code, sol_wd=None, stu_wd=None, **kwargs):
    with ChDir(sol_wd or os.getcwd()):
        sol_process, _, _ = run_single_process(pec, sol_code, **kwargs)

    with ChDir(stu_wd or os.getcwd()):
        stu_process, raw_stu_output, error = run_single_process(pec, stu_code, **kwargs)

    return sol_process, stu_process, raw_stu_output, error


# todo:
#  imports from local modules (solution needs to be materialised somewhere)
#  converge with xbackend (pythonbackend + look at scalabackend)
#  move towards xwhat controlling all execution and xbackend providing the execution interface?
# running with arbitrary wd + path + flags (now only wd) needed?
#  e.g. `python -m project.run
#  allow setting env vars? e.g. PYTHONPATH, could help running more complex setup
#  allow prepending code? set_env? e.g. (automatically) setting __file__?
def run(state, relative_working_dir=None, solution_dir="../solution", run_solution=True):
    """Run the focused student and solution code in the specified location

    This function can be used after ``check_file`` to execute student and solution code.
    The arguments allow configuring the correct context for execution.

    SCT functions chained after this one that execute pieces of code (custom expressions or the focused part of a file)
    execute in the same student and solution locations as the file.

    .. note::

        This function does not execute the file itself, but code in memory.
        This can have an impact when:

        - the solution code imports from a different file in the expected solution (code that is not installed)
        - using functionality depending on e.g. ``__file__`` and ``inspect``

        When the expected code has imports from a different file that is part of the exercise,
        it can only work if the solution code provided earlier does not have these imports but instead
        has all that functionality inlined.

    Args:
        relative_working_dir (str): if specified, this relative path is the subdirectory
            inside the student and solution context in which the code is executed
        solution_dir (str): a relative path, ``solution`` by default,
            that sets the root of the solution context, relative to that of the student execution context
        state (State): state as passed by the SCT chain. Don't specify this explicitly.

    If ``relative_working_dir`` is not set, it will be the directory the file was loaded from by ``check_file``
    and fall back to the root of the student execution context (the working directory pythonwhat runs in).

    The ``solution_dir`` helps to prevent solution side effects from conflicting with those of the student.
    If the set or derived value of ``relative_working_dir`` is an absolute path,
    ``relative_working_dir`` will not be used to form the solution execution working directory:
    the solution code will be executed in the root of the solution execution context.

    :Example:

        Suppose the student and solution have a file ``script.py`` in ``/home/repl/``::

            if True:
                a = 1

            print("Hi!")

        We can check it with this SCT (with ``file_content`` containing the expected file content)::

            Ex().check_file(
                "script.py",
                solution_code=file_content
            ).run().multi(
                check_object("a").has_equal_value(),
                has_printout(0)
            )
    """
    # todo:
    # look into executing the file itself
    # and keeping the process alive to extract values
    if relative_working_dir is None:
        if getattr(state, "path", False):
            relative_working_dir = state.path.parent
        else:
            relative_working_dir = ""

    if not os.path.isabs(str(relative_working_dir)):
        sol_wd = Path(os.getcwd(), solution_dir, relative_working_dir)
    else:
        sol_wd = Path(os.getcwd(), solution_dir)

    os.makedirs(str(sol_wd), exist_ok=True)
    stu_wd = Path(os.getcwd(), relative_working_dir)
    sol_code = state.solution_code or "" if run_solution else ""

    sol_process, stu_process, raw_stu_output, error = run_exercise(
        pec="",
        sol_code=sol_code,
        stu_code=state.student_code,
        sol_wd=sol_wd,
        stu_wd=stu_wd,
    )
    return state.to_child(
        student_process=stu_process,
        solution_process=sol_process,
        raw_student_output=raw_stu_output,
        reporter=Reporter(state.reporter, errors=[error] if error else []),
    )



from tcs_pythonwhat import utils
import dill
import pickle
import tcs_pythonwhat
import ast
import inspect
from copy import deepcopy
from pickle import PicklingError
from tcs_pythonwhat.utils_env import set_context_vals, assign_from_ast
from contextlib import contextmanager
from functools import partial, wraps
from protowhat.failure import InstructorError


# Shell is passed as a parameter to partially applied functions in executeTask
# Process is passed as a parameter in SCT function


def process_task(f):
    """Decorator to (optionally) run function in a process."""
    sig = inspect.signature(f)

    @wraps(f)
    def wrapper(*args, **kwargs):
        # get bound arguments for call
        ba = sig.bind_partial(*args, **kwargs)
        # when process is specified, remove from args and use to execute
        process = ba.arguments.get("process")
        if process:
            ba.arguments["process"] = None
            # partial function since shell argument may have been left
            # unspecified, as it will be passed when the process executes
            pf = partial(wrapper, *ba.args, **ba.kwargs)
            return process.executeTask(pf)
        # otherwise, run original function
        return f(*ba.args, **ba.kwargs)

    return wrapper


def get_env(ns):
    if "__env__" in ns:
        return ns["__env__"]
    else:
        return ns


@contextmanager
def capture_output():
    import sys
    from io import StringIO

    oldout, olderr = sys.stdout, sys.stderr
    out = [StringIO(), StringIO()]
    sys.stdout, sys.stderr = out
    yield out
    sys.stdout, sys.stderr = oldout, olderr
    out[0] = out[0].getvalue()
    out[1] = out[1].getvalue()


# MC
@process_task
def getOptionFromProcess(process, name, shell):
    return shell.user_ns[name]


# Is a variable is defined in the process?
@process_task
def isDefinedInProcess(name, process, shell):
    return name in get_env(shell.user_ns)


# Is a variable is of a certain class in the process?
@process_task
def isInstanceInProcess(name, klass, process, shell):
    return isinstance(get_env(shell.user_ns)[name], klass)


# Get the columns of a Pandas data frame in the process
@process_task
def getColumnsInProcess(name, process, shell):
    return list(get_env(shell.user_ns)[name].columns)


# Is a key defined in a collection in the process?
@process_task
def isDefinedCollInProcess(name, key, process, shell):
    return key in get_env(shell.user_ns)[name]


# Get the signature of a function inside the process


def get_signature(name, mapped_name, signature, manual_sigs, env):
    if isinstance(signature, str):
        if signature in manual_sigs:
            signature = inspect.Signature(manual_sigs[signature])
        else:
            raise InstructorError.from_message("signature error - указанная сигнатура не найдена.")

    if signature is None:
        # establish function
        try:
            fun = eval(mapped_name, env)
        except:
            raise InstructorError.from_message("%s() не было найдено." % mapped_name)

        # first go through manual sigs
        # try to get signature
        try:
            if name in manual_sigs:
                signature = inspect.Signature(manual_sigs[name])
            else:
                # it might be a method, and we have to find the general method name
                if "." in mapped_name:
                    els = name.split(".")
                    try:
                        els[0] = type(eval(els[0], env)).__name__
                        generic_name = ".".join(els[:])
                    except:
                        raise InstructorError.from_message("signature error - не удается преобразовать вызов")
                    if generic_name in manual_sigs:
                        signature = inspect.Signature(manual_sigs[generic_name])
                    else:
                        raise InstructorError.from_message(
                            "signature error - %s не встроена" % generic_name
                        )
                else:
                    raise InstructorError.from_message("manual signature not found")
        except Exception as e:
            try:
                signature = inspect.signature(fun)
            except:
                raise InstructorError.from_message(e.args[0] + " и не удается определить сигнатуру")

    return signature


# Get the signature of a function based on an object inside the process
@process_task
def getSignatureInProcess(name, mapped_name, signature, manual_sigs, process, shell):
    return get_signature(
        name=name,
        mapped_name=mapped_name,
        signature=signature,
        manual_sigs=manual_sigs,
        env=get_env(shell.user_ns),
    )


@process_task
def getSignatureFromObjInProcess(obj_char, process, shell):
    try:
        return inspect.signature(eval(obj_char, get_env(shell.user_ns)))
    except:
        return None


# Stuff for test_with

from contextlib import ExitStack


def context_env_update(context_list, env):
    es = ExitStack()
    for item in context_list:
        # create context manager and enter
        tmp_name = "__pw_cm"
        cm_code = compile(ast.Expression(item.context_expr), "<context_eval>", "eval")
        env[tmp_name] = es.enter_context(eval(cm_code, env))

        # assign to its optional_vars in separte dict
        if item.optional_vars:
            code = assign_from_ast(item.optional_vars, tmp_name)
            exec(code, env)

    return es


@process_task
def setUpNewEnvInProcess(context, process, shell):
    shell.user_ns["__env__"] = utils.copy_env(shell.user_ns)
    try:
        es = context_env_update(context, shell.user_ns["__env__"])
        shell.user_ns["__exit_stack__"] = es
        return True
    except Exception as e:
        return e


# break down environment
def context_objs_exit(es):
    try:
        es.close()
        return True
    except Exception as e:
        return e


@process_task
def breakDownNewEnvInProcess(process, shell):
    try:
        res = context_objs_exit(shell.user_ns["__exit_stack__"])
        del shell.user_ns["__exit_stack__"]
        del shell.user_ns["__env__"]
        return res
    except:
        return False


# Tasks that may need to serialize across processes ===========================

# Get a bytes or string representation of an object in the process
@process_task
def getClass(name, process, shell):
    try:
        obj = get_env(shell.user_ns)[name]
        obj_type = type(obj)
        return obj_type.__module__ + "." + obj_type.__name__
    except:
        return None


@process_task
def convert(name, converter, process, shell):
    return dill.loads(converter)(get_env(shell.user_ns)[name])


@process_task
def getStreamPickle(name, process, shell):
    try:
        return pickle.dumps(get_env(shell.user_ns)[name])
    except:
        return None


@process_task
def getStreamDill(name, process, shell):
    try:
        return dill.dumps(get_env(shell.user_ns)[name])
    except:
        return None


class ReprFail:
    def __init__(self, info):
        self.info = info


def getRepresentation(name, process):
    obj_class = getClass(name, process)
    converters = pythonwhat.State.State.root_state.converters
    if obj_class in converters:
        repres = convert(name, dill.dumps(converters[obj_class]), process)
        if errored(repres):
            return ReprFail("manual conversion failed: {}".format(repres))
        else:
            return repres
    else:
        # first try to pickle
        try:
            stream = getStreamPickle(name, process)
            if not errored(stream):
                return pickle.loads(stream)
        except:
            pass

        # if it failed, try to dill
        try:
            stream = getStreamDill(name, process)
            if not errored(stream):
                return dill.loads(stream)
            return ReprFail(
                "dilling inside process failed for %s - write manual converter"
                % obj_class
            )
        except PicklingError:
            return ReprFail(
                "undilling of bytestream failed with PicklingError - write manual converter"
            )
        except Exception as e:
            return ReprFail(
                "undilling of bytestream failed for class %s - write manual converter."
                "Error: %s - %s" % (obj_class, type(e), e)
            )


def errored(el):
    return el is None or (isinstance(el, list) and "backend-error" in str(el))


# Make wrapper for getting an object representation from process --------------
class UndefinedValue:
    pass


def getResultFromProcess(res, tempname, process):
    """Get a value from process, return tuple of value, res if succesful"""
    if not isinstance(res, (UndefinedValue, Exception)):
        value = getRepresentation(tempname, process)
        return value, res
    else:
        return res, str(res)


# decorator to automatically get value after running process task function
def get_rep(f):
    sig = inspect.signature(f)

    @wraps(f)
    def wrapper(*args, **kwargs):
        # get bound arguments for call
        ba = sig.bind_partial(*args, **kwargs)
        ba.apply_defaults()
        # get tempname, process arg values
        tempname = ba.arguments["tempname"]
        process = ba.arguments["process"]
        # run process task
        res = f(*args, **kwargs)
        # get result from task
        return getResultFromProcess(res, tempname, process)

    return wrapper


## Get the output of a tree (with setting envs, pre_code and/er expr_code)
@process_task
def get_output(f, process, shell, *args, **kwargs):
    with capture_output() as out:
        res = f(*args, process=process, shell=shell, **kwargs)

    out_str = out[0].strip()
    if not isinstance(res, Exception):
        toret = out_str or "no printouts"
        return toret, toret
    else:
        return res, str(res)


@process_task
def get_error(f, *args, **kwargs):
    res = f(*args, **kwargs)
    return (res, str(res)) if isinstance(res, Exception) else (None, res)


# General tasks to eval or exec code, with decorated counterparts -------------


@process_task
def taskRunEval(
    tree,
    process,
    shell,
    env=None,
    extra_env=None,
    context=None,
    context_vals=None,
    pre_code="",
    expr_code="",
    name="",
    copy=True,
    tempname="_evaluation_object_",
    call=None,
):
    """
    Eval an expression tree (with setting envs, pre_code and/or expr_code)
    Utility function later wrapped to extract either result (end state of a variable), output (stdout) or error

    Args:
        tree (ast): current focused ast, used to get code to execute
        process: manages shell (see local.py)
        shell: link to to get process namespace from execution up until now
        env: update value in focused code by name
        extra_env: variables to be replaced in focused code by name from extra_env in has_expr
        context: sum of set_context in sct chain
        context_vals: extra context argument in has_expr
        pre_code: argument in has_expr to execute code before evaluating, for example to set a seed
        expr_code: code to execute instead of focused code
        name: extract value after executing focused expr_code (~post_code)
        copy: copy entire env because our expr_code could have side effects
        tempname: key for the result when it is added to context, only for v1 sct's
        call: only used in v1 sct's

    Returns:
        str: output of the executed code
    """
    try:
        # Prepare code and mode -----------------------------------------------
        # Verify if expr_code is expression code (returning a value) or just runnable code.
        if (  # expr_code returns nothing and then we will extract a value
            expr_code and name
        ) or (  # No expr_code and the tree is of a node type that does not evaluate to have output
            not expr_code and isinstance(tree, ast.Module)
        ):
            # We are not focused on an expression (no output)
            mode = "exec"
        else:
            mode = "eval"
            # Wrap the focused node in the tree so it can be run with eval()
            if not isinstance(tree, (ast.Module, ast.Expression, ast.Expr)):
                tree = ast.Expression(tree)

        # Expression code takes precedence over tree code
        if expr_code:
            code = expr_code
            tree = ast.parse(code, mode=mode)
        else:  # Compile the tree to Python code
            code = compile(tree, "<script>", mode)

        # Set up environment --------------------------------------------------
        # Unpack 'container nodes' before checking if a deepcopy is needed
        if isinstance(tree, ast.Module):
            tree = tree.body
        if isinstance(tree, ast.Expression):
            tree = tree.body
        if isinstance(tree, ast.Expr):
            tree = tree.value

        # Avoid a deep copy if specified or if the ast node type indicates we are looking up a variable by name
        # ast.Name, ast.Subscript and ast.Load most of the time do not have side effects in the environment,
        #   making a deepcopy unnecessary
        if not copy or (
            isinstance(tree, (ast.Name, ast.Subscript))
            and isinstance(tree.ctx, ast.Load)
        ):
            new_env = dict(get_env(shell.user_ns))  # shallow copy of env
        else:
            # might raise an error if object refuses pickle interface
            # used by deepcopy to restore class
            new_env = utils.copy_env(get_env(shell.user_ns))

        # Apply additional env and context variables
        if env is not None:
            new_env.update(deepcopy(env))
        if extra_env is not None:
            new_env.update(deepcopy(extra_env))
        if context is not None:
            set_context_vals(new_env, context, context_vals)

        # Execute code --------------------------------------------------------
        # Run pre_code if specified
        if pre_code:
            exec(pre_code, new_env)

        if mode == "eval":
            obj = eval(code, new_env)
        else:
            exec(code, new_env)
            obj = "exec only"

        # If name given, get from new_env
        if name:
            try:
                obj = eval(name, new_env)
            except NameError:
                return UndefinedValue()

        # Backwards compatibility with v1 SCT's
        # If object is callable, args and kwargs can be passed in to be used in a call on object
        if call is not None:
            obj = obj(*call["args"], **call["kwargs"])

        # Set object as temp variable in original environment, so we can
        # later get its class, etc.., in order to extract it from process
        get_env(shell.user_ns)[tempname] = obj

        return str(obj)

    except Exception as e:
        return e


getResultInProcess = get_rep(taskRunEval)
getOutputInProcess = partial(get_output, taskRunEval)
getErrorInProcess = partial(get_error, taskRunEval)



from typing import Dict

from protowhat.Feedback import Feedback as ProtoFeedback


class Feedback(ProtoFeedback):
    ast_highlight_offset = {"column_start": 1}

    @classmethod
    def get_highlight_position(cls, highlight) -> Dict[str, int]:
        if getattr(highlight, "first_token", None) and getattr(
            highlight, "last_token", None
        ):
            return {
                "line_start": highlight.first_token.start[0],
                "column_start": highlight.first_token.start[1],
                "line_end": highlight.last_token.end[0],
                "column_end": highlight.last_token.end[1],
            }



from functools import partial

from protowhat.sct_syntax import link_to_state
from tcs_pythonwhat.checks.check_function import check_function
from protowhat.failure import TestFail, InstructorError
from tcs_pythonwhat.checks.check_funcs import check_args
from tcs_pythonwhat.checks.has_funcs import has_equal_value, has_equal_ast, has_printout

# this is done by the chain for v2
# it's only needed when a new state is created and (possibly) used elsewhere
check_function = link_to_state(check_function)


def arg_test(state, name, do_eval, missing_msg, incorrect_msg):
    arg_state = check_args(state, name=name, missing_msg=missing_msg)

    append = incorrect_msg is None

    if isinstance(do_eval, bool):
        if do_eval:
            has_equal_value(
                arg_state, incorrect_msg=incorrect_msg, append=append, copy=False
            )
        else:
            has_equal_ast(arg_state, incorrect_msg=incorrect_msg, append=append)


def test_function(
    state,
    name,
    index=1,
    args=None,
    keywords=None,
    eq_condition="equal",
    do_eval=True,
    not_called_msg=None,
    args_not_specified_msg=None,
    incorrect_msg=None,
    add_more=False,
    **kwargs
):
    index = index - 1

    # if root-level (not in compound statement) calls: use has_printout
    if name == "print" and state.parent_state is None and do_eval:
        try:
            return has_printout(state, index=index, not_printed_msg=incorrect_msg)
        except TestFail:
            # The test didn't pass; just continue with the more strict check_function test.
            pass

    fun_state = check_function(
        state, name=name, index=index, missing_msg=not_called_msg, signature=False
    )

    if args is None:
        args = [
            k
            for k, value in fun_state.solution_parts["args"].items()
            if isinstance(k, int)
        ]

    if keywords is None:
        keywords = [
            k
            for k, value in fun_state.solution_parts["args"].items()
            if isinstance(k, str)
        ]

    arg_test_partial = partial(
        arg_test,
        fun_state,
        do_eval=do_eval,
        missing_msg=args_not_specified_msg,
        incorrect_msg=incorrect_msg,
    )

    [arg_test_partial(name=i) for i in range(len(args))]
    [arg_test_partial(name=keyword) for keyword in keywords]

    return state


def test_function_v2(
    state,
    name,
    index=1,
    params=[],
    signature=True,
    eq_condition="equal",
    do_eval=True,
    not_called_msg=None,
    params_not_matched_msg=None,
    params_not_specified_msg=None,
    incorrect_msg=None,
    add_more=False,
    **kwargs
):

    index = index - 1

    if not isinstance(params, list):
        raise InstructorError.from_message(
            "Inside test_function_v2, make sure to specify a LIST of params."
        )

    if isinstance(do_eval, bool) or do_eval is None:
        do_eval = [do_eval] * len(params)

    if len(params) != len(do_eval):
        raise InstructorError.from_message(
            "Inside test_function_v2, make sure that do_eval has the same length as params."
        )

    # if params_not_specified_msg is a str or None, convert into list
    if isinstance(params_not_specified_msg, str) or params_not_specified_msg is None:
        params_not_specified_msg = [params_not_specified_msg] * len(params)

    if len(params) != len(params_not_specified_msg):
        raise InstructorError.from_message(
            "Inside test_function_v2, make sure that params_not_specified_msg has the same length as params."
        )

    # if incorrect_msg is a str or None, convert into list
    if isinstance(incorrect_msg, str) or incorrect_msg is None:
        incorrect_msg = [incorrect_msg] * len(params)

    if len(params) != len(incorrect_msg):
        raise InstructorError.from_message(
            "Inside test_function_v2, make sure that incorrect_msg has the same length as params."
        )

    # if root-level (not in compound statement) calls that can be evaluated: use has_printout
    eligible = do_eval[0] if isinstance(do_eval, list) and len(do_eval) > 0 else do_eval
    if name == "print" and state.parent_state is None and eligible:
        try:
            return has_printout(state, index=index, not_printed_msg=incorrect_msg[0])
        except TestFail:
            # The test didn't pass; just continue with the more strict check_function test.
            pass

    if len(params) == 0:
        signature = False

    fun_state = check_function(
        state,
        name=name,
        index=index,
        missing_msg=not_called_msg,
        params_not_matched_msg=params_not_matched_msg,
        signature=signature,
    )

    for i in range(len(params)):
        arg_test(
            fun_state,
            name=params[i],
            do_eval=do_eval[i],
            missing_msg=params_not_specified_msg[i],
            incorrect_msg=incorrect_msg[i],
        )

    return state



from types import ModuleType
import copy
import os


def include_v1():
    return os.environ.get("PYTHONWHAT_V2_ONLY", "") != "1"


def v2_only():
    return not include_v1()

def shorten_string(text):
    if len(text) > 50:
        text = text[0:45] + "..."
    return text

def has_newline(text):
    return "\n" in text


def copy_env(env):
    mutableTypes = (tuple, list, dict)
    # One list comprehension to filter list. Might need some cleaning, but it
    # works
    ipy_ignore = ["In", "Out", "get_ipython", "quit", "exit"]
    update_env = {
        key: copy.deepcopy(value)
        for key, value in env.items()
        if not any(
            (key.startswith("_"), isinstance(value, ModuleType), key in ipy_ignore)
        )
        and isinstance(value, mutableTypes)
    }
    updated_env = dict(env)
    updated_env.update(update_env)
    return updated_env


def first_lower(s):
    return s[:1].lower() + s[1:] if s else ""


def check_str(x):
    assert isinstance(x, str), "объект не является строкой там, где ожидается строка."
    return x


def check_dict(x):
    assert isinstance(x, dict), "объект не является словарем там, где ожидается словарь."
    return x


def check_process(x):
    assert "Process" in x.__class__.__name__
    return x
